"""
:mod:`textacy.text_stats.api`: Compute basic and readability statistics of documents.
"""
import functools
import logging
from typing import Optional, Tuple

import pyphen
from cachetools import cached
from cachetools.keys import hashkey
from spacy.tokens import Doc

from .. import cache, extract
from . import basics, readability


LOGGER = logging.getLogger(__name__)


class TextStats:
    """
    Class to compute a variety of basic and readability statistics for a given doc,
    where each stat is a lazily-computed attribute.

    .. code-block:: pycon

        >>> text = next(textacy.datasets.CapitolWords().texts(limit=1))
        >>> doc = textacy.make_spacy_doc(text)
        >>> ts = textacy.text_stats.TextStats(doc)
        >>> ts.n_words
        136
        >>> ts.n_unique_words
        80
        >>> ts.entropy
        6.00420319027642
        >>> ts.flesch_kincaid_grade_level
        11.817647058823532
        >>> ts.flesch_reading_ease
        50.707745098039254

    Some stats vary by language or are designed for use with specific languages:

    .. code-block:: pycon

        >>> text = (
        ...     "Muchos años después, frente al pelotón de fusilamiento, "
        ...     "el coronel Aureliano Buendía había de recordar aquella tarde remota "
        ...     "en que su padre lo llevó a conocer el hielo."
        ... )
        >>> doc = textacy.make_spacy_doc(text, lang="es")
        >>> ts = textacy.text_stats.TextStats(doc)
        >>> ts.n_words
        28
        >>> ts.perspicuity_index
        56.46000000000002
        >>> ts.mu_legibility_index
        71.18644067796609

    Each of these stats have stand-alone functions in :mod:`textacy.text_stats.basics`
    and :mod:`textacy.text_stats.readability` with more detailed info and links
    in the docstrings -- when in doubt, read the docs!

    Args:
        doc: A text document tokenized and (optionally) sentence-segmented by spaCy.
    """

    def __init__(self, doc: Doc):
        self.doc = doc
        self.lang = doc.vocab.lang
        self.words = tuple(
            extract.words(doc, filter_punct=True, filter_stops=False, filter_nums=False)
        )
        self._n_sents: Optional[int] = None
        self._n_words: Optional[int] = None
        self._n_unique_words: Optional[int] = None
        self._n_long_words: Optional[int] = None
        self._n_chars_per_word: Optional[Tuple[int, ...]] = None
        self._n_chars: Optional[int] = None
        self._n_syllables_per_word: Optional[Tuple[int, ...]] = None
        self._n_syllables: Optional[int] = None
        self._n_monosyllable_words: Optional[int] = None
        self._n_polysyllable_words: Optional[int] = None
        self._entropy: Optional[float] = None

    @property
    def n_sents(self) -> int:
        """
        Number of sentences in document.

        See Also:
            :func:`textacy.text_stats.basics.n_sents()`
        """
        if self._n_sents is None:
            self._n_sents = basics.n_sents(self.doc)
        return self._n_sents

    @property
    def n_words(self) -> int:
        """
        Number of words in document.

        See Also:
            :func:`textacy.text_stats.basics.n_words()`
        """
        if self._n_words is None:
            self._n_words = basics.n_words(self.words)
        return self._n_words

    @property
    def n_unique_words(self) -> int:
        """
        Number of *unique* words in document.

        See Also:
            :func:`textacy.text_stats.basics.n_unique_words()`
        """
        if self._n_unique_words is None:
            self._n_unique_words = basics.n_unique_words(self.words)
        return self._n_unique_words

    @property
    def n_long_words(self) -> int:
        """
        Number of long words in document.

        See Also:
            :func:`textacy.text_stats.basics.n_long_words()`
        """
        # TODO: should we vary char threshold by lang?
        if self._n_long_words is None:
            self._n_long_words = basics.n_long_words(
                self.n_chars_per_word, min_n_chars=7,
            )
        return self._n_long_words

    @property
    def n_chars_per_word(self) -> Tuple[int, ...]:
        """
        Number of characters for each word in document.

        See Also:
            :func:`textacy.text_stats.basics.n_chars_per_word()`
        """
        if self._n_chars_per_word is None:
            self._n_chars_per_word = basics.n_chars_per_word(self.words)
        return self._n_chars_per_word

    @property
    def n_chars(self) -> int:
        """
        Total number of characters in document.

        See Also:
            :func:`textacy.text_stats.basics.n_chars()`
        """
        if self._n_chars is None:
            self._n_chars = basics.n_chars(self.n_chars_per_word)
        return self._n_chars

    @property
    def n_syllables_per_word(self) -> Tuple[int, ...]:
        """
        Number of syllables for each word in document.

        See Also:
            :func:`textacy.text_stats.basics.n_syllables_per_word()`
        """
        if self._n_syllables_per_word is None:
            self._n_syllables_per_word = basics.n_syllables_per_word(
                self.words, self.lang,
            )
        return self._n_syllables_per_word

    @property
    def n_syllables(self) -> int:
        """
        Total number of syllables in document.

        See Also:
            :func:`textacy.text_stats.basics.n_syllables()`
        """
        if self._n_syllables is None:
            self._n_syllables = basics.n_syllables(self.n_syllables_per_word)
        return self._n_syllables

    @property
    def n_monosyllable_words(self) -> int:
        """
        Number of monosyllobic words in document.

        See Also:
            :func:`textacy.text_stats.basics.n_monosyllable_words()`
        """
        if self._n_monosyllable_words is None:
            self._n_monosyllable_words = basics.n_monosyllable_words(
                self.n_syllables_per_word,
            )
        return self._n_monosyllable_words

    @property
    def n_polysyllable_words(self) -> int:
        """
        Number of polysyllobic words in document.

        See Also:
            :func:`textacy.text_stats.basics.n_polysyllable_words()`
        """
        # TODO: should we vary syllable threshold by lang?
        if self._n_polysyllable_words is None:
            self._n_polysyllable_words = basics.n_polysyllable_words(
                self.n_syllables_per_word, min_n_syllables=3,
            )
        return self._n_polysyllable_words

    @property
    def entropy(self) -> float:
        """
        Entropy of words in document.

        See Also:
            :func:`textacy.text_stats.basics.entropy()`
        """
        if self._entropy is None:
            self._entropy = basics.entropy(self.words)
        return self._entropy

    @property
    def automated_readability_index(self) -> float:
        """
        Readability test for English-language texts. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.automated_readability_index()`
        """
        return readability.automated_readability_index(
            self.n_chars, self.n_words, self.n_sents,
        )

    @property
    def automatic_arabic_readability_index(self) -> float:
        """
        Readability test for Arabic-language texts. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.automatic_arabic_readability_index()`
        """
        if self.lang != "ar":
            LOGGER.warning(
                "doc lang = '%s', but automatic arabic readability index is meant "
                "for use on Arabic-language texts, only"
            )
        return readability.automatic_arabic_readability_index(
            self.n_chars, self.n_words, self.n_sents,
        )

    @property
    def coleman_liau_index(self) -> float:
        """
        Readability test, not language-specific. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.coleman_liau_index()`
        """
        return readability.coleman_liau_index(self.n_chars, self.n_words, self.n_sents)

    @property
    def flesch_kincaid_grade_level(self) -> float:
        """
        Readability test, not language-specific. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.flesch_kincaid_grade_level()`
        """
        return readability.flesch_kincaid_grade_level(
            self.n_syllables, self.n_words, self.n_sents,
        )

    @property
    def flesch_reading_ease(self) -> float:
        """
        Readability test with several language-specific formulations.
        Higher value => easier text.

        See Also:
            :func:`textacy.text_stats.readability.flesch_reading_ease()`
        """
        return readability.flesch_reading_ease(
            self.n_syllables, self.n_words, self.n_sents, lang=self.lang
        )

    @property
    def gulpease_index(self) -> float:
        """
        Readability test for Italian-language texts. Higher value => easier text.

        See Also:
            :func:`textacy.text_stats.readability.gulpease_index()`
        """
        if self.lang != "it":
            LOGGER.warning(
                "doc lang = '%s', but gulpease index is meant for use on "
                "Italian-language texts, only"
            )
        return readability.gulpease_index(self.n_chars, self.n_words, self.n_sents)

    @property
    def gunning_fog_index(self) -> float:
        """
        Readability test, not language-specific. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.gunning_fog_index()`
        """
        return readability.gunning_fog_index(
            self.n_words, self.n_polysyllable_words, self.n_sents,
        )

    @property
    def lix(self) -> float:
        """
        Readability test for both English- and non-English-language texts.
        Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.lix()`
        """
        return readability.lix(self.n_words, self.n_long_words, self.n_sents)

    @property
    def mu_legibility_index(self) -> float:
        """
        Readability test for Spanish-language texts. Higher value => easier text.

        See Also:
            :func:`textacy.text_stats.readability.mu_legibility_index()`
        """
        if self.lang != "es":
            LOGGER.warning(
                "doc lang = '%s', but mu legibility index is meant for use on "
                "Spanish-language texts, only"
            )
        return readability.mu_legibility_index(self.n_chars_per_word)

    @property
    def perspicuity_index(self) -> float:
        """
        Readability test for Spanish-language texts. Higher value => easier text.

        See Also:
            :func:`textacy.text_stats.readability.perspicuity_index()`
        """
        if self.lang != "es":
            LOGGER.warning(
                "doc lang = '%s', but perspicuity index is meant for use on "
                "Spanish-language texts, only"
            )
        return readability.perspicuity_index(
            self.n_syllables, self.n_words, self.n_sents,
        )

    @property
    def smog_index(self) -> float:
        """
        Readability test, not language-specific. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.smog_index()`
        """
        return readability.smog_index(self.n_polysyllable_words, self.n_sents)

    @property
    def wiener_sachtextformel(self) -> float:
        """
        Readability test for German-language texts. Higher value => more difficult text.

        See Also:
            :func:`textacy.text_stats.readability.wiener_sachtextformel()`
        """
        if self.lang != "es":
            LOGGER.warning(
                "doc lang = '%s', but wiener sachtextformel is meant for use on "
                "German-language texts, only"
            )
        return readability.wiener_sachtextformel(
            self.n_words,
            self.n_polysyllable_words,
            self.n_monosyllable_words,
            self.n_long_words,
            self.n_sents,
            variant=1,
        )


@cached(cache.LRU_CACHE, key=functools.partial(hashkey, "hyphenator"))
def load_hyphenator(lang: str):
    """
    Load an object that hyphenates words at valid points, as used in LaTex typesetting.

    Args:
        lang: Standard 2-letter language abbreviation. To get a list of valid values::

            >>> import pyphen; pyphen.LANGUAGES

    Returns:
        :class:`pyphen.Pyphen()`
    """
    LOGGER.debug("loading '%s' language hyphenator", lang)
    return pyphen.Pyphen(lang=lang)



from __future__ import annotations

from typing import Callable, List, Optional, Sequence, Tuple

import numpy as np
import thinc
from cytoolz import itertoolz
from thinc.api import Model, chain, concatenate


def get_model_preds(model: Model, texts: List[str], classes: np.ndarray) -> List[str]:
    """
    Get model predictions for multiple texts as class labels rather than as a 2dim
    matrix of prediction probabilities.
    """
    # predict in batches, otherwise memory blows UP
    results = (
        result
        for texts_pt in itertoolz.partition_all(1000, texts)
        for result in get_topn_preds_and_probs(model.predict(texts_pt), 1, classes)
    )
    return [lang for result in results for lang, _ in result]


def get_topn_preds_and_probs(
    preds: np.ndarray,
    topn: int,
    classes: np.ndarray,
) -> List[List[Tuple[str, float]]]:
    # TODO
    # if only need 1 (max) value, use faster numpy ops?
    # if topn == 1:
    #     idxs = np.argmax(preds, axis=1)
    #     pred_probs = np.max(preds, axis=1)
    #     pred_langs = self.classes[idxs]
    #     return list(zip(pred_langs, pred_probs))
    # otherwise, do the full array sorts to get topn max
    # else:
    idxs = np.argsort(preds, axis=1)[:, ::-1][:, :topn]
    pred_probs = np.sort(preds, axis=1)[:, ::-1][:, :topn]
    pred_langs = classes[idxs]
    return [
        list(zip(pred_langs[i], pred_probs[i]))
        for i in range(pred_probs.shape[0])
    ]


def LangIdentifierModelV2(
    ns: Sequence[int] = (1, 2, 3),
    embed_dim: int = 100,
    hidden_width: int = 512,
    dropout: Optional[float] = 0.1,
) -> Model[List[str], thinc.types.Floats2d]:
    """
    Build a language identification model inspired by Google's CLD3.

    Args:
        ns: Set of "n" for which character "n"-grams are extracted from input texts.
            If 1, only unigrams (single characters) are used; if [1, 2], then both
            unigrams and bigrams are used; and so on.
        embed_dim: Size of the vectors into which each set of ngrams are embedded.
        hidden_width: Width of the dense layer with Relu activation, just before
            the final prediction (Softmax) layer.
        dropout: Dropout rate to avoid overfitting.

    Returns:
        Thinc :class:`Model`.
    """
    with Model.define_operators({">>": chain}):
        model = (
            MultiCharNgramsEmbedding(
                ns=list(ns),
                max_chars=1000,
                lower=True,
                num_vectors=[2000 * n for n in ns],
                embed_dims=embed_dim,
                dropout=dropout,
            )
            >> thinc.layers.Relu(
                nI=embed_dim * len(ns),
                nO=hidden_width,
                dropout=dropout,
            )
            >> thinc.layers.Softmax(nI=hidden_width)
        )
    return model


def MultiCharNgramsEmbedding(
    ns: List[int],
    max_chars: int,
    lower: bool,
    num_vectors: int | List[int],
    embed_dims: int | List[int],
    dropout: Optional[float],
) -> Model[List[str], thinc.types.Floats1d]:
    """
    Args:
        ns
        max_chars
        lower
        num_vectors
        embed_dims
        dropout
    """
    numn = len(ns)
    num_vectors = [num_vectors] * numn if isinstance(num_vectors, int) else num_vectors
    embed_dims = [embed_dims] * numn if isinstance(embed_dims, int) else embed_dims
    with Model.define_operators({">>": chain}):
        model = concatenate(
            *[
                CharNgramsEmbedding(
                    n=n,
                    max_chars=max_chars,
                    lower=lower,
                    num_vectors=nvec,
                    embed_dim=edim,
                    dropout=dropout,
                )
                for n, nvec, edim in zip(ns, num_vectors, embed_dims)
            ]
        )
    return model


def CharNgramsEmbedding(
    n: int,
    max_chars: int,
    lower: bool,
    num_vectors: int,
    embed_dim: int,
    dropout: Optional[float],
) -> Model[List[str], thinc.types.Floats1d]:
    """
    Args:
        n
        max_chars
        lower
        num_vectors
        embed_dim
        dropout
    """
    with Model.define_operators({">>": chain}):
        model = (
            text_to_char_ngrams(n, max_chars, lower)
            >> thinc.layers.strings2arrays()
            >> thinc.layers.with_array(
                thinc.layers.HashEmbed(
                    nO=embed_dim,
                    nV=num_vectors,
                    dropout=dropout,
                    column=0,
                )
            )
            >> thinc.layers.list2ragged()
            >> thinc.layers.reduce_mean()
        )
    return model


def text_to_char_ngrams(
    n: int,
    max_chars: int,
    lower: bool,
) -> Model[List[str], List[List[str]]]:
    """
    Custom data type transfer thinc layer that transforms a sequence of text strings
    into a sequence of sequence of character ngram strings. Like this::

        ["a short text.", "another text."] => [["a ", " s", "sh", "ho", ...], ...]

    Args:
        n: Number of adjacent characters to combine into an ngram.
        max_chars: Max number of characters from the start of the text to transform
            into character ngrams.
        lower: If True, lowercase text before extracting character ngrams; otherwise,
            leave text casing as-is.
    """

    def forward(
        model: Model, texts: List[str], is_train: bool
    ) -> Tuple[List[List[str]], Callable]:
        if lower is True:
            texts = (text[:max_chars].lower() for text in texts)
        else:
            texts = (text[:max_chars] for text in texts)
        if n == 1:
            char_ngs = [list(text) for text in texts]
        else:
            char_ngs = [
                [text[i : i + n] for i in range(len(text) - n + 1)]
                for text in texts
            ]

        def backprop(dY):
            return []

        return (char_ngs, backprop)

    return Model(
        "texts_to_char_ngrams",
        forward,
        attrs={"n": n, "max_chars": max_chars, "lower": lower},
    )



"""
Sequence-based Metrics
----------------------

:mod:`textacy.similarity.sequences`: Normalized similarity metrics built on
sequence-based algorithms that identify and measure the subsequences common to each.
"""
import difflib
from typing import Sequence


def matching_subsequences_ratio(
    seq1: Sequence[str], seq2: Sequence[str], **kwargs
) -> float:
    """
    Measure the similarity between two sequences of strings by finding
    contiguous matching subsequences without any "junk" elements and normalizing
    by the total number of elements.

    Args:
        seq1
        seq2
        **kwargs
            isjunk: Optional[Callable[str], bool] = None
            autojunk: bool = True

    Returns:
        Similarity between ``seq1`` and ``seq2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar sequences of strings

    Reference:
        https://docs.python.org/3/library/difflib.html#difflib.SequenceMatcher.ratio
    """
    return difflib.SequenceMatcher(a=seq1, b=seq2, **kwargs).ratio()



"""
:mod:`textacy.io.http`: Functions for reading data from URLs via streaming HTTP requests
and either reading it into memory or writing it directly to disk.
"""
from __future__ import annotations

import logging
from contextlib import closing
from typing import Iterable, Optional, Tuple

import requests
from tqdm import tqdm

from .. import types, utils
from . import utils as io_utils

LOGGER = logging.getLogger(__name__)


def read_http_stream(
    url: str,
    *,
    lines: bool = False,
    decode_unicode: bool = False,
    chunk_size: int = 1024,
    auth: Optional[Tuple[str, str]] = None,
) -> Iterable[str] | Iterable[bytes]:
    """
    Read data from ``url`` in a stream, either all at once or line-by-line.

    Args:
        url: URL to which a GET request is made for data.
        lines: If False, yield all of the data at once; otherwise, yield data line-by-line.
        decode_unicode: If True, yield data as unicode, where the encoding
            is taken from the HTTP response headers; otherwise, yield bytes.
        chunk_size: Number of bytes read into memory per chunk.
            Because decoding may occur, this is not necessarily the length of each chunk.
        auth: (username, password) pair for simple HTTP authentication required (if at all)
            to access the data at ``url``.

            .. seealso:: http://docs.python-requests.org/en/master/user/authentication/

    Yields:
        If ``lines`` is True, the next line in the response data,
        which is bytes if ``decode_unicode`` is False or unicode otherwise.
        If ``lines`` is False, yields the full response content, either as bytes
        or unicode.
    """
    # always close the connection
    with closing(requests.get(url, stream=True, auth=auth)) as r:
        # set fallback encoding if unable to infer from headers
        if r.encoding is None:
            r.encoding = "utf-8"
        if lines is False:
            if decode_unicode is True:
                yield r.text
            else:
                yield r.content
        else:
            lines_ = r.iter_lines(chunk_size=chunk_size, decode_unicode=decode_unicode)
            for line in lines_:
                if line:
                    yield line


def write_http_stream(
    url: str,
    filepath: types.PathLike,
    *,
    mode: str = "wt",
    encoding: Optional[str] = None,
    make_dirs: bool = False,
    chunk_size: int = 1024,
    auth: Optional[Tuple[str, str]] = None,
) -> None:
    """
    Download data from ``url`` in a stream, and write successive chunks
    to disk at ``filepath``.

    Args:
        url: URL to which a GET request is made for data.
        filepath: Path to file on disk to which data will be written.
        mode: Mode with which ``filepath`` is opened.
        encoding: Name of the encoding used to decode or encode the data
            in ``filepath``. Only applicable in text mode.

            .. note:: The encoding on the HTTP response is inferred from its
               headers, or set to 'utf-8' as a fall-back in the case that no
               encoding is detected. It is *not* set by ``encoding``.

        make_dirs: If True, automatically create (sub)directories if
            not already present in order to write ``filepath``.
        chunk_size: Number of bytes read into memory per chunk.
            Because decoding may occur, this is not necessarily the length of each chunk.
        auth: (username, password) pair for simple HTTP authentication required (if at all)
            to access the data at ``url``.

            .. seealso:: http://docs.python-requests.org/en/master/user/authentication/
    """
    decode_unicode = True if "t" in mode else False
    filepath = utils.to_path(filepath).resolve()
    if make_dirs is True:
        io_utils._make_dirs(filepath, mode)
    # use `closing` to ensure connection and progress bar *always* close
    with closing(requests.get(url, stream=True, auth=auth)) as r:
        LOGGER.info("downloading data from %s ...", url)
        # set fallback encoding if unable to infer from headers
        if r.encoding is None:
            r.encoding = "utf-8"
        total = int(r.headers.get("content-length", 0))
        with closing(tqdm(unit="B", unit_scale=True, total=total)) as pbar:
            with filepath.open(mode=mode, encoding=encoding) as f:
                chunks = r.iter_content(
                    chunk_size=chunk_size, decode_unicode=decode_unicode
                )
                for chunk in chunks:
                    # needed (?) to filter out "keep-alive" new chunks
                    if chunk:
                        pbar.update(len(chunk))
                        f.write(chunk)



"""
Pipeline
--------

:mod:`textacy.preprocessing.pipeline`: Basic functionality for composing multiple
preprocessing steps into a single callable pipeline.
"""
from typing import Callable

from cytoolz import functoolz


def make_pipeline(*funcs: Callable[[str], str]) -> Callable[[str], str]:
    """
    Make a callable pipeline that takes a text as input, passes it through one or more
    functions in sequential order, then outputs a single (preprocessed) text string.

    This function is intended as a lightweight convenience for users, allowing them
    to flexibly specify which (and in which order) preprocessing functions are to be
    applied to raw texts, then treating the whole thing as a single callable.

    .. code-block:: pycon

        >>> from textacy import preprocessing
        >>> preproc = preprocessing.make_pipeline(
        ...     preprocessing.replace.hashtags,
        ...     preprocessing.replace.user_handles,
        ...     preprocessing.replace.emojis,
        ... )
        >>> preproc("@spacy_io is OSS for industrial-strength NLP in Python developed by @explosion_ai 💥")
        '_USER_ is OSS for industrial-strength NLP in Python developed by _USER_ _EMOJI_'
        >>> preproc("hacking with my buddy Isaac Mewton 🥰 #PawProgramming")
        'hacking with my buddy Isaac Mewton _EMOJI_ _TAG_'

    Args:
        *funcs

    Returns:
        Pipeline composed of ``*funcs`` that applies each in sequential order.
    """
    return functoolz.compose_left(*funcs)



"""
:mod:`textacy.spacier.core`: Convenient entry point for loading spaCy language pipelines
and making spaCy docs.
"""
from __future__ import annotations

import functools
import logging
import pathlib
from typing import Optional

import spacy
from cachetools import cached
from cachetools.keys import hashkey
from spacy.language import Language
from spacy.tokens import Doc

from . import utils as sputils
from .. import cache, errors, types, utils


LOGGER = logging.getLogger(__name__)


@cached(cache.LRU_CACHE, key=functools.partial(hashkey, "spacy_lang"))
def load_spacy_lang(name: str | pathlib.Path, **kwargs) -> Language:
    """
    Load a spaCy ``Language`` — a shared vocabulary and language-specific data
    for tokenizing text, and (if available) model data and a processing pipeline
    containing a sequence of components for annotating a document — and cache results,
    for quick reloading as needed.

    Note that as of spaCy v3, for which pipeline aliases are no longer allowed,
    this function is just a convenient access point to underlying :func:`spacy.load()`.

    .. code-block:: pycon

        >>> en_nlp = textacy.load_spacy_lang("en_core_web_sm")
        >>> en_nlp = textacy.load_spacy_lang("en_core_web_sm", disable=("parser",))
        >>> textacy.load_spacy_lang("ar")
        ...
        OSError: [E050] Can't find model 'ar'. It doesn't seem to be a Python package or a valid path to a data directory.

    Args:
        name: Name or path to the spaCy language pipeline to load.
        **kwargs

    Note:
        Although spaCy's API specifies some kwargs as ``List[str]``, here we require
        ``Tuple[str, ...]`` equivalents. Language pipelines are stored in an LRU cache
        with unique identifiers generated from the hash of the function name and args —
        and lists aren't hashable.

    Returns:
        Loaded spaCy ``Language``.

    Raises:
        OSError

    See Also:
        https://spacy.io/api/top-level#spacy.load
    """
    spacy_lang = spacy.load(name, **kwargs)
    LOGGER.info("loaded '%s' spaCy language pipeline", name)
    return spacy_lang


def make_spacy_doc(
    data: types.DocData,
    lang: types.LangLikeInContext,
    *,
    chunk_size: Optional[int] = None,
) -> Doc:
    """
    Make a :class:`spacy.tokens.Doc` from valid inputs, and automatically
    load/validate :class:`spacy.language.Language` pipelines to process ``data``.

    Make a ``Doc`` from text:

    .. code-block:: pycon

        >>> text = "To be, or not to be, that is the question."
        >>> doc = make_spacy_doc(text, "en_core_web_sm")
        >>> doc._.preview
        'Doc(13 tokens: "To be, or not to be, that is the question.")'

    Make a ``Doc`` from a (text, metadata) pair, aka a "record":

    .. code-block:: pycon

        >>> record = (text, {"author": "Shakespeare, William"})
        >>> doc = make_spacy_doc(record, "en_core_web_sm")
        >>> doc._.preview
        'Doc(13 tokens: "To be, or not to be, that is the question.")'
        >>> doc._.meta
        {'author': 'Shakespeare, William'}

    Specify the language pipeline used to process the text in a few different ways:

    .. code-block:: pycon

        >>> make_spacy_doc(text, lang="en_core_web_sm")
        >>> make_spacy_doc(text, lang=textacy.load_spacy_lang("en_core_web_sm"))
        >>> make_spacy_doc(text, lang=lambda txt: "en_core_web_sm")

    Ensure that an already-processed ``Doc`` is compatible with ``lang``:

    .. code-block:: pycon

        >>> spacy_lang = textacy.load_spacy_lang("en_core_web_sm")
        >>> doc = spacy_lang(text)
        >>> make_spacy_doc(doc, lang="en_core_web_sm")
        >>> make_spacy_doc(doc, lang="es_core_news_sm")
        ...
        ValueError: `spacy.Vocab` used to process document must be the same as that used by the `lang` pipeline ('es_core_news_sm')

    Args:
        data: Make a :class:`spacy.tokens.Doc` from a text or (text, metadata) pair.
            If already a ``Doc``, ensure that it's compatible with ``lang``
            to avoid surprises downstream, and return it as-is.
        lang: Language with which spaCy processes (or processed) ``data``,
            represented as the full name of a spaCy language pipeline, the path on disk
            to it, an already instantiated pipeline, or a callable function that takes
            the text component of ``data`` and outputs one of the above representations.
        chunk_size: Size of chunks in number of characters into which ``text`` will be
            split before processing each via spaCy and concatenating the results
            into a single ``Doc``.

            .. note:: This is intended as a workaround for processing very long texts,
               for which spaCy is unable to allocate enough RAM. For best performance,
               chunk size should be somewhere between 1e3 and 1e7 characters,
               depending on how much RAM you have available.

               Since chunking is done by *character*, chunks' boundaries likely
               won't respect natural language segmentation, and as a result
               spaCy's models may make mistakes on sentences/words that cross them.

    Returns:
        Processed spaCy Doc.

    Raises:
        TypeError
        ValueError
    """
    if isinstance(data, str):
        return _make_spacy_doc_from_text(data, lang, chunk_size)
    elif isinstance(data, Doc):
        return _make_spacy_doc_from_doc(data, lang)
    elif utils.is_record(data):
        return _make_spacy_doc_from_record(data, lang, chunk_size)
    else:
        raise TypeError(errors.type_invalid_msg("data", type(data), types.DocData))


def _make_spacy_doc_from_text(
    text: str, lang: types.LangLikeInContext, chunk_size: Optional[int]
) -> Doc:
    spacy_lang = sputils.resolve_langlikeincontext(text, lang)
    if chunk_size:
        doc = _make_spacy_doc_from_text_chunks(text, spacy_lang, chunk_size)
    else:
        doc = spacy_lang(text)
    return doc


def _make_spacy_doc_from_record(
    record: types.Record, lang: types.LangLikeInContext, chunk_size: Optional[int]
) -> Doc:
    text, meta = record
    spacy_lang = sputils.resolve_langlikeincontext(text, lang)
    if chunk_size:
        doc = _make_spacy_doc_from_text_chunks(text, spacy_lang, chunk_size)
    else:
        doc = spacy_lang(text)
    doc._.meta = meta
    return doc


def _make_spacy_doc_from_text_chunks(text: str, lang: Language, chunk_size: int) -> Doc:
    text_chunks = (text[i : i + chunk_size] for i in range(0, len(text), chunk_size))
    return Doc.from_docs(list(lang.pipe(text_chunks)))


def _make_spacy_doc_from_doc(doc: Doc, lang: types.LangLikeInContext) -> Doc:
    spacy_lang = sputils.resolve_langlikeincontext(doc.text, lang)
    # we want to make sure that the language used to create `doc` is the same as
    # the one passed here; however, the best we can do (bc of spaCy's API) is ensure
    # that they share the same vocab
    if doc.vocab is not spacy_lang.vocab:
        raise ValueError(
            f"`spacy.Vocab` used to process document ({doc.vocab}) must be the same "
            f"as that used by the `lang` pipeline ({spacy_lang.vocab})"
        )
    return doc



"""
:mod:`textacy.errors`: Helper functions for making consistent errors.
"""
from typing import Any, Collection


def value_invalid_msg(name: str, value: Any, valid_values: Collection[Any]) -> str:
    return f"`{name}` value = {value} is invalid; value must be one of {valid_values}."


def type_invalid_msg(name: str, val_type, valid_val_type) -> str:
    return f"`{name}` type = {val_type} is invalid; type must match {valid_val_type}."



"""
:mod:`textacy.io.matrix`: Functions for reading from and writing to disk CSC and CSR
sparse matrices in numpy binary format.
"""
from __future__ import annotations

from typing import Union

import numpy as np
import scipy.sparse as sp

from .. import errors, types, utils
from . import utils as io_utils


def read_sparse_matrix(
    filepath: types.PathLike, *, kind: str = "csc",
) -> sp.csc_matrix | sp.csr_matrix:
    """
    Read the data, indices, indptr, and shape arrays from a ``.npz`` file on disk
    at ``filepath``, and return an instantiated sparse matrix.

    Args:
        filepath: Path to file on disk from which data will be read.
        kind ({'csc', 'csr'}): Kind of sparse matrix to instantiate.

    Returns:
        An instantiated sparse matrix, whose type depends on the value of ``kind``.

    See Also:
        https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.io.html#numpy-binary-files-npy-npz
    """
    npz_file = np.load(filepath)
    if kind == "csc":
        return sp.csc_matrix(
            (npz_file["data"], npz_file["indices"], npz_file["indptr"]),
            shape=npz_file["shape"],
        )
    elif kind == "csr":
        return sp.csr_matrix(
            (npz_file["data"], npz_file["indices"], npz_file["indptr"]),
            shape=npz_file["shape"],
        )
    else:
        raise ValueError(errors.value_invalid_msg("kind", kind, {"csc", "csr"}))


def write_sparse_matrix(
    data: sp.csc_matrix | sp.csr_matrix,
    filepath: types.PathLike,
    *,
    compressed: bool = True,
    make_dirs: bool = False,
) -> None:
    """
    Write sparse matrix ``data`` to disk at ``filepath``, optionally compressed,
    into a single ``.npz`` file.

    Args:
        data
        filepath: Path to file on disk to which data will be written. If ``filepath``
            does not end in ``.npz``, that extension is automatically appended to the name.
        compressed: If True, save arrays into a single file in compressed numpy binary format.
        make_dirs: If True, automatically create (sub)directories
            if not already present in order to write ``filepath``.

    See Also:
        https://docs.scipy.org/doc/numpy-1.13.0/reference/routines.io.html#numpy-binary-files-npy-npz
    """
    if not isinstance(data, (sp.csc_matrix, sp.csr_matrix)):
        raise TypeError(
            errors.type_invalid_msg(
                "data", type(data), Union[sp.csc_matrix, sp.csr_matrix]
            )
        )
    filepath = utils.to_path(filepath).resolve()
    if make_dirs is True:
        io_utils._make_dirs(filepath, "w")
    if compressed is True:
        np.savez_compressed(
            str(filepath),
            data=data.data,
            indices=data.indices,
            indptr=data.indptr,
            shape=data.shape,
        )
    else:
        np.savez(
            str(filepath),
            data=data.data,
            indices=data.indices,
            indptr=data.indptr,
            shape=data.shape,
        )



import datetime
import os

import pytest

from textacy.datasets.imdb import IMDB

DATASET = IMDB()


def _skipif():
    try:
        DATASET._check_data()
        return False
    except OSError:
        return True


pytestmark = pytest.mark.skipif(
    _skipif(),
    reason="IMDB dataset must be downloaded before running tests",
)


@pytest.mark.skip("No need to download a new dataset every time")
def test_download(tmpdir):
    dataset = IMDB(data_dir=str(tmpdir))
    dataset.download()
    assert dataset.data_dir.joinpath("aclImdb").is_dir()


def test_oserror(tmpdir):
    dataset = IMDB(data_dir=str(tmpdir))
    with pytest.raises(OSError):
        _ = list(dataset.texts())


def test_texts():
    texts = list(DATASET.texts(limit=3))
    assert len(texts) > 0
    for text in texts:
        assert isinstance(text, str)


def test_texts_limit():
    for limit in (1, 5, 10):
        assert sum(1 for _ in DATASET.texts(limit=limit)) == limit


def test_texts_min_len():
    for min_len in (100, 200, 500):
        assert all(
            len(text) >= min_len for text in DATASET.texts(min_len=min_len, limit=10)
        )


def test_records():
    for text, meta in DATASET.records(limit=3):
        assert isinstance(text, str)
        assert isinstance(meta, dict)


def test_records_subset():
    subsets = ({"train"}, {"test", "train"})
    for subset in subsets:
        records = list(DATASET.records(subset=subset, limit=10))
        assert all(
            meta["subset"] in subset
            for text, meta in records
        )


def test_records_label():
    labels = ({"pos"}, {"neg", "unsup"})
    for label in labels:
        records = list(DATASET.records(label=label, limit=10))
        assert all(
            meta["label"] in label
            for text, meta in records
        )


def test_records_rating_range():
    rating_ranges = ((6, 8), [8, 10])
    for rating_range in rating_ranges:
        records = list(DATASET.records(rating_range=rating_range, limit=10))
        assert all(
            rating_range[0] <= meta["rating"] < rating_range[1]
            for text, meta in records
        )


def test_bad_filters():
    bad_filters = (
        {"min_len": -1},
    )
    for bad_filter in bad_filters:
        with pytest.raises(ValueError):
            list(DATASET.texts(**bad_filter))
    bad_filters = (
        {"rating_range": 5},
        {"rating_range": ("1", "10")},
    )
    for bad_filter in bad_filters:
        with pytest.raises(TypeError):
            list(DATASET.texts(**bad_filter))



import pytest

import textacy
from textacy import datasets
from textacy.extract import keyterms as kt


DATASET = datasets.CapitolWords()

pytestmark = pytest.mark.skipif(
    DATASET.filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def spacy_doc():
    text = next(DATASET.texts(min_len=1500, limit=1))
    return textacy.make_spacy_doc(text, lang="en_core_web_sm")


@pytest.fixture(scope="module")
def empty_spacy_doc():
    return textacy.make_spacy_doc("", lang="en_core_web_sm")


def test_default(spacy_doc):
    result = kt.sgrank(spacy_doc)
    assert isinstance(result, list) and len(result) > 0
    assert all(isinstance(ts, tuple) and len(ts) == 2 for ts in result)
    assert all(
        isinstance(ts[0], str) and isinstance(ts[1], float)
        for ts in result
    )


def test_ngrams_1(spacy_doc):
    result = kt.sgrank(spacy_doc, ngrams=1)
    assert len(result) > 0
    assert all(len(term.split()) == 1 for term, _ in result)


def test_ngrams_2_3(spacy_doc):
    result = kt.sgrank(spacy_doc, ngrams=(2, 3))
    assert len(result) > 0
    assert all(2 <= len(term.split()) <= 3 for term, _ in result)


def test_topn(spacy_doc):
    for n in (5, 25):
        result = kt.sgrank(spacy_doc, topn=n)
        assert 0 < len(result) <= n


def test_topn_float(spacy_doc):
    result = kt.sgrank(spacy_doc, topn=0.2)
    assert len(result) > 0
    with pytest.raises(ValueError):
        _ = kt.sgrank(spacy_doc, topn=2.0)


def test_normalize_lower(spacy_doc):
    result = kt.sgrank(spacy_doc, normalize="lower")
    assert len(result) > 0
    assert all(term == term.lower() for term, _ in result)


def test_normalize_none(spacy_doc):
    result = kt.sgrank(spacy_doc, normalize=None)
    assert len(result) > 0
    assert any(term != term.lower() for term, _ in result)


def test_normalize_callable(spacy_doc):
    result = kt.sgrank(spacy_doc, normalize=lambda tok: tok.text.upper())
    assert len(result) > 0
    assert all(term == term.upper() for term, _ in result)


def test_window_size(spacy_doc):
    result_10 = kt.sgrank(spacy_doc, window_size=10)
    result_100 = kt.sgrank(spacy_doc, window_size=100)
    assert len(result_10) > 0 and len(result_100) > 0
    assert result_10 != result_100
    with pytest.raises(ValueError):
        _ = kt.sgrank(spacy_doc, window_size=1)


def test_empty_doc(empty_spacy_doc):
    result = kt.sgrank(empty_spacy_doc)
    assert isinstance(result, list)
    assert len(result) == 0



from . import edits, hybrid, sequences, tokens
from .edits import hamming, levenshtein, jaro, character_ngrams
from .hybrid import token_sort_ratio, monge_elkan
from .sequences import matching_subsequences_ratio
from .tokens import jaccard, sorensen_dice, tversky, cosine, bag



from __future__ import annotations

from typing import Callable, Collection, Iterable, List, Optional, Tuple, Union

from spacy.tokens import Doc, Span

from .. import errors, types, utils


class CharNgramsTokenizer:
    """
    Transform each text in a sequence of texts into a sequence of character ngram tokens
    as strings, where ngrams for multiple "n"s are concatenated together for each text.

    Args:
        ns: Number of contiguous characters to include in each token.
            If an int, only ``ns`` n-grams are included as tokens; if a Collection[int],
            then n-grams for each n in ``ns`` are concatenated together per text.
        pad: If True, pad texts by adding ``n - 1`` underscore characters to each side;
            if False, leave text as-is.
        normalize: Optional callable or string alias for a callable that processes
            each text before transforming it into a sequence of character ngrams.
            If None, text is left as-is; if "lower", text is lowercased; otherwise,
            the callable is applied directly to each text.
    """

    def __init__(
        self,
        ns: int | Collection[int],
        *,
        pad: bool = False,
        normalize: Optional[str | Callable[[str], str]] = None,
    ):
        self.ns = utils.to_collection(ns, int, tuple)
        self.pad = pad
        self.normalize = self._init_normalize(normalize)

    def __str__(self) -> str:
        return (
            "CharNgramsTokenizer("
            f"ns={self.ns}, pad={self.pad}, normalize={self.normalize}"
            ")"
        )

    def _init_normalize(
        self, normalize: Optional[str | Callable[[str], str]]
    ) -> Optional[Callable[[str], str]]:
        if normalize is None:
            return None
        elif callable(normalize):
            return normalize
        elif normalize == "lower":
            return lambda text: text.lower()
        else:
            raise ValueError()

    def fit(self, textables: Iterable[str | types.DocLike]) -> "CharNgramsTokenizer":
        return self

    def transform(
        self,
        textables: Iterable[str | types.DocLike],
    ) -> Iterable[Tuple[str, ...]]:
        """
        Convert a sequence of texts or textable spaCy objects (Doc or Span)
        into an ordered, nested sequence of character ngrams.

        Args:
            textables

        Yields:
            Ordered sequence of character ngrams as strings for next textable.
        """
        texts = (self._to_text(textable) for textable in textables)
        if self.normalize is not None:
            texts = (self.normalize(text) for text in texts)
        for text in texts:
            tokens: List[str] = []
            for n in self.ns:
                if self.pad is True and n > 1:
                    pad_chars = "_" * (n - 1)
                    text = f"{pad_chars}{text}{pad_chars}"
                tokens.extend(text[i : i + n] for i in range(len(text) - n + 1))
            yield tuple(tokens)

    def _to_text(self, textable: str | types.DocLike) -> str:
        if isinstance(textable, str):
            return textable
        elif isinstance(textable, (Doc, Span)):
            return textable.text
        else:
            raise TypeError(
                errors.type_invalid_msg("textable", textable, Union[str, types.DocLike])
            )



import os
import re

import pytest

from textacy.datasets import wikimedia


WIKINEWS = wikimedia.Wikinews(lang="en", version="current")
WIKIPEDIA = wikimedia.Wikipedia(lang="en", version="current")


@pytest.mark.skipif(
    WIKIPEDIA.filepath is None,
    reason="Wikinews dataset must be downloaded before running tests",
)
class TestWikipedia:

    @pytest.mark.skip("No need to download a new dataset every time")
    def test_download(self, tmpdir):
        dataset = wikimedia.Wikipedia(data_dir=str(tmpdir))
        dataset.download()
        assert os.path.isfile(dataset.filepath)

    def test_oserror(self, tmpdir):
        dataset = wikimedia.Wikipedia(data_dir=str(tmpdir))
        with pytest.raises(OSError):
            _ = list(dataset.texts())

    def test_texts(self):
        texts = list(WIKIPEDIA.texts(limit=3))
        assert len(texts) > 0
        for text in texts:
            assert isinstance(text, str)

    def test_texts_limit(self):
        for limit in (1, 5, 10):
            assert sum(1 for _ in WIKIPEDIA.texts(limit=limit)) == limit

    def test_texts_min_len(self):
        for min_len in (100, 200, 500):
            assert all(
                len(text) >= min_len
                for text in WIKIPEDIA.texts(min_len=min_len, limit=10)
            )

    def test_records(self):
        for text, meta in WIKIPEDIA.records(limit=3):
            assert isinstance(text, str)
            assert isinstance(meta, dict)

    def test_records_limit(self):
        for limit in (1, 5, 10):
            assert sum(1 for _ in WIKIPEDIA.records(limit=limit)) == limit

    def test_records_min_len(self):
        for min_len in (100, 200, 500):
            assert all(
                len(text) >= min_len
                for text, meta in WIKIPEDIA.records(min_len=min_len, limit=10)
            )

    def test_records_category(self):
        categories = (
            {"Living people"},
            {"All stub articles", "Coordinates on Wikidata"},
        )
        for category in categories:
            records = list(WIKIPEDIA.records(category=category, limit=3))
            assert any(
                ctgry in meta["categories"]
                for _, meta in records
                for ctgry in category
            )

    def test_records_wiki_link(self):
        wiki_links = (
            {"United_States"},
            {"France", "England"},
        )
        for wiki_link in wiki_links:
            records = list(WIKIPEDIA.records(wiki_link=wiki_link, limit=3))
            assert any(
                wl in meta["wiki_links"]
                for _, meta in records
                for wl in wiki_link
            )


@pytest.mark.skipif(
    WIKINEWS.filepath is None,
    reason="Wikinews dataset must be downloaded before running tests",
)
class TestWikinews:

    @pytest.mark.skip("No need to download a new dataset every time")
    def test_download(self, tmpdir):
        dataset = wikimedia.Wikinews(data_dir=str(tmpdir))
        dataset.download()
        assert os.path.isfile(dataset.filepath)

    def test_oserror(self, tmpdir):
        dataset = wikimedia.Wikinews(data_dir=str(tmpdir))
        with pytest.raises(OSError):
            _ = list(dataset.texts())

    def test_texts(self):
        texts = list(WIKINEWS.texts(limit=3))
        assert len(texts) > 0
        for text in texts:
            assert isinstance(text, str)

    def test_texts_limit(self):
        for limit in (1, 5, 10):
            assert sum(1 for _ in WIKINEWS.texts(limit=limit)) == limit

    def test_texts_min_len(self):
        for min_len in (100, 200, 500):
            assert all(
                len(text) >= min_len
                for text in WIKINEWS.texts(min_len=min_len, limit=10)
            )

    def test_records(self):
        for text, meta in WIKINEWS.records(limit=3):
            assert isinstance(text, str)
            assert isinstance(meta, dict)

    def test_records_limit(self):
        for limit in (1, 5, 10):
            assert sum(1 for _ in WIKINEWS.records(limit=limit)) == limit

    def test_records_min_len(self):
        for min_len in (100, 200, 500):
            assert all(
                len(text) >= min_len
                for text, meta in WIKINEWS.records(min_len=min_len, limit=10)
            )

    def test_records_category(self):
        categories = (
            {"Politics and conflicts"},
            {"United States", "North America"},
        )
        for category in categories:
            records = list(WIKINEWS.records(category=category, limit=3))
            assert any(
                ctgry in meta["categories"]
                for _, meta in records
                for ctgry in category
            )

    def test_records_wiki_link(self):
        wiki_links = (
            {"Reuters"},
            {"United_States", "BBC_News"},
        )
        for wiki_link in wiki_links:
            records = list(WIKINEWS.records(wiki_link=wiki_link, limit=3))
            assert any(
                wl in meta["wiki_links"]
                for _, meta in records
                for wl in wiki_link
            )



"""
Acronyms
--------

:mod:`textacy.extract.acronyms`: Extract acronyms and their definitions from a document
or sentence through rule-based pattern-matching of the annotated tokens.
"""
from __future__ import annotations

import collections
from operator import itemgetter
from typing import Dict, Iterable, List, Optional, Set, Tuple

import numpy as np
from spacy.tokens import Span, Token

from .. import constants, types


def acronyms(doclike: types.DocLike) -> Iterable[Token]:
    """
    Extract tokens whose text is "acronym-like" from a document or sentence,
    in order of appearance.

    Args:
        doclike

    Yields:
        Next acronym-like ``Token``.
    """
    for tok in doclike:
        if is_acronym(tok.text):
            yield tok


def acronyms_and_definitions(
    doclike: types.DocLike,
    known_acro_defs: Optional[Dict[str, str]] = None,
) -> Dict[str, List[str]]:
    """
    Extract a collection of acronyms and their most likely definitions, if available,
    from a spacy-parsed doc. If multiple definitions are found for a given acronym,
    only the most frequently occurring definition is returned.

    Args:
        doclike
        known_acro_defs: If certain acronym/definition pairs
            are known, pass them in as {acronym (str): definition (str)};
            algorithm will not attempt to find new definitions

    Returns:
        Unique acronyms (keys) with matched definitions (values)

    References:
        Taghva, Kazem, and Jeff Gilbreth. "Recognizing acronyms and their definitions."
        International Journal on Document Analysis and Recognition 1.4 (1999): 191-198.
    """
    # process function arguments
    acro_defs: Dict[str, List[Tuple[str, float]]] = collections.defaultdict(list)
    if not known_acro_defs:
        known_acronyms = set()
    else:
        for acro, def_ in known_acro_defs.items():
            acro_defs[acro] = [(def_, 1.0)]
        known_acronyms = set(acro_defs.keys())

    if isinstance(doclike, Span):
        sents = [doclike]
    else:  # spacy.Doc
        sents = doclike.sents

    # iterate over sentences and their tokens
    for sent in sents:
        max_ind = len(sent) - 1

        for i, token in enumerate(sent):

            token_ = token.text
            if token_ in known_acronyms or is_acronym(token_) is False:
                continue

            # define definition search window(s)
            window_size = min(2 * len(token_), len(token_) + 5)
            windows = [
                sent[max(i - window_size, 0) : i],
                sent[min(i + 1, max_ind) : min(i + window_size + 1, max_ind)],
            ]
            # if candidate inside (X) or -X-, only look in pre-window
            if 0 < i < max_ind:
                adjacent_tokens = sent[i - 1].text + sent[i + 1].text
                if adjacent_tokens in {"()", "--", "––"}:
                    windows.pop()

            # iterate over possible windows
            # filtering for valid definition strings
            for window in windows:
                window_ = window.text
                # window text can't be all uppercase
                if window_.isupper():
                    continue
                # window can't contain separating punctuation
                if "!" in window_ or "?" in window_ or ":" in window_ or ";" in window_:
                    continue
                # acronym definition can't contain itself: no ouroboros!
                if token_ in window_:
                    continue
                # window must contain at least one character used in acronym
                if not any(char in window_ for char in token_):
                    continue
                definition, confidence = _get_acronym_definition(
                    token_, window, threshold=0.8
                )
                if definition:
                    acro_defs[token_].append((definition, confidence))

            if not acro_defs.get(token_):
                acro_defs[token_].append(("", 0.0))

    # vote by confidence score in the case of multiple definitions
    for acro, defs in acro_defs.items():
        if len(defs) == 1:
            acro_defs[acro] = defs[0][0]
        else:
            acro_defs[acro] = sorted(defs, key=itemgetter(1), reverse=True)[0][0]

    return dict(acro_defs)


def _get_acronym_definition(
    acronym: str, window: Span, threshold: float = 0.8,
) -> Tuple[str, float]:
    """
    Identify most likely definition for an acronym given a list of tokens.

    Args:
        acronym: acronym for which definition is sought
        window: a span of tokens from which definition extraction will be attempted
        threshold: minimum "confidence" in definition required for acceptance;
            valid values in [0.0, 1.0]; higher value => stricter threshold

    Returns:
        Most likely definition for given acronym ('' if none found),
        along with the confidence assigned to it

    References:
        Taghva, Kazem, and Jeff Gilbreth. "Recognizing acronyms and their definitions."
        International Journal on Document Analysis and Recognition 1.4 (1999): 191-198.
    """

    def build_lcs_matrix(X, Y):
        m = len(X)
        n = len(Y)
        b = np.zeros((m, n), dtype=int)
        c = np.zeros((m, n), dtype=int)
        for i in range(0, m):
            for j in range(0, n):
                if X[i] == Y[j]:
                    c[i, j] = c[i - 1, j - 1] + 1
                    b[i, j] = 1
                elif c[i - 1, j] >= c[i, j - 1]:
                    c[i, j] = c[i - 1, j]
                else:
                    c[i, j] = c[i, j - 1]
        return c, b

    def parse_lcs_matrix(b, start_i, start_j, lcs_length, stack, vectors):
        m = b.shape[0]
        n = b.shape[1]
        for i in range(start_i, m):
            for j in range(start_j, n):
                if b[i, j] == 1:
                    s = (i, j)
                    stack.append(s)
                    if lcs_length == 1:
                        vec = [np.NaN] * n
                        for k, l in stack:
                            vec[l] = k
                        vectors.append(vec)
                    else:
                        parse_lcs_matrix(b, i + 1, j + 1, lcs_length - 1, stack, vectors)
                    stack = []
        return vectors

    def vector_values(v, types):
        vv = {}
        first = v.index(int(np.nanmin(v)))
        last = v.index(int(np.nanmax(v)))
        vv["size"] = (last - first) + 1
        vv["distance"] = len(v) - last
        vv["stop_count"] = 0
        vv["misses"] = 0
        for i in range(first, last + 1):
            if v[i] >= 0 and types[i] == "s":
                vv["stop_count"] += 1
            elif v[i] is None and types[i] not in ["s", "h"]:
                vv["misses"] += 1
        return vv

    def compare_vectors(A, B, types):
        vv_A = vector_values(A, types)
        vv_B = vector_values(B, types)
        # no one-letter matches, sorryboutit
        if vv_A["size"] == 1:
            return B
        elif vv_B["size"] == 1:
            return A
        if vv_A["misses"] > vv_B["misses"]:
            return B
        elif vv_A["misses"] < vv_B["misses"]:
            return A
        if vv_A["stop_count"] > vv_B["stop_count"]:
            return B
        if vv_A["stop_count"] < vv_B["stop_count"]:
            return A
        if vv_A["distance"] > vv_B["distance"]:
            return B
        elif vv_A["distance"] < vv_B["distance"]:
            return A
        if vv_A["size"] > vv_B["size"]:
            return B
        elif vv_A["size"] < vv_B["size"]:
            return A
        return A

    # get definition window's leading characters and word types
    def_leads = []
    def_types = []
    for tok in window:
        tok_text = tok.text
        if tok.is_stop:
            def_leads.append(tok_text[0])
            def_types.append("s")
        elif is_acronym(tok_text):
            def_leads.append(tok_text[0])
            def_types.append("a")
        elif "-" in tok_text and not tok_text.startswith("-"):
            tok_split = [t[0] for t in tok_text.split("-") if t]
            def_leads.extend(tok_split)
            def_types.extend("H" if i == 0 else "h" for i in range(len(tok_split)))
        else:
            def_leads.append(tok_text[0])
            def_types.append("w")
    def_leads = "".join(def_leads).lower()
    def_types = "".join(def_types)

    # extract alphanumeric characters from acronym
    acr_leads = "".join(c for c in acronym if c.isalnum())
    # handle special cases of '&' and trailing 's'
    acr_leads = acr_leads.replace("&", "a")
    if acr_leads.endswith("s"):
        # bail out if it's only a 2-letter acronym to start with, e.g. 'Is'
        if len(acr_leads) == 2:
            return ("", 0)
        acr_leads = acr_leads[:-1]
    acr_leads = acr_leads.lower()

    c, b = build_lcs_matrix(acr_leads, def_leads)

    # 4.4.1
    lcs_length = c[c.shape[0] - 1, c.shape[1] - 1]
    confidence = lcs_length / len(acronym)
    if confidence < threshold:
        return ("", confidence)

    vecs = parse_lcs_matrix(b, 0, 0, lcs_length, [], [])
    # first letter of acronym must be present
    vecs = [vec for vec in vecs if 0 in vec]
    if not vecs:
        return ("", confidence)

    best_vec = vecs[0]
    for vec in vecs[1:]:
        best_vec = compare_vectors(best_vec, vec, def_types)

    first = best_vec.index(int(np.nanmin(best_vec)))
    last = best_vec.index(int(np.nanmax(best_vec)))

    definition = window[first : last + 1].text
    if len(definition.split()) == 1:
        return ("", confidence)

    return (definition, confidence)


def is_acronym(token: str, exclude: Optional[Set[str]] = None) -> bool:
    """
    Pass single token as a string, return True/False if is/is not valid acronym.

    Args:
        token: Single word to check for acronym-ness
        exclude: If technically valid but not actual acronyms are known in advance,
            pass them in as a set of strings; matching tokens will return False.

    Returns:
        Whether or not ``token`` is an acronym.
    """
    # exclude certain valid acronyms from consideration
    if exclude and token in exclude:
        return False
    # don't allow empty strings
    if not token:
        return False
    # don't allow spaces
    if " " in token:
        return False
    # 2-character acronyms can't have lower-case letters
    if len(token) == 2 and not token.isupper():
        return False
    # acronyms can't be all digits
    if token.isdigit():
        return False
    # acronyms must have at least one upper-case letter or start/end with a digit
    if not any(char.isupper() for char in token) and not (
        token[0].isdigit() or token[-1].isdigit()
    ):
        return False
    # acronyms must have between 2 and 10 alphanumeric characters
    if not 2 <= sum(1 for char in token if char.isalnum()) <= 10:
        return False
    # only certain combinations of letters, digits, and '&/.-' allowed
    if not constants.RE_ACRONYM.match(token):
        return False
    return True



from .acros import acronyms, acronyms_and_definitions
from .basics import words, ngrams, entities, noun_chunks, terms
from .kwic import keyword_in_context
from .matches import token_matches, regex_matches
from .triples import (
    direct_quotations, semistructured_statements, subject_verb_object_triples
)
from .utils import terms_to_strings, clean_term_strings, aggregate_term_variants

from .extensions import get_doc_extensions, remove_doc_extensions, set_doc_extensions

set_doc_extensions()



import pytest
import spacy

from textacy import load_spacy_lang, make_spacy_doc


# TODO: why are we using such a long text here??
TEXT = (
    "Since the so-called \"statistical revolution\" in the late 1980s and mid 1990s, "
    "much Natural Language Processing research has relied heavily on machine learning. "
    "Formerly, many language-processing tasks typically involved the direct hand coding "
    "of rules, which is not in general robust to natural language variation. "
    "The machine-learning paradigm calls instead for using statistical inference "
    "to automatically learn such rules through the analysis of large corpora of typical "
    "real-world examples (a corpus is a set of documents, possibly with human or "
    "computer annotations). Many different classes of machine learning algorithms "
    "have been applied to NLP tasks. These algorithms take as input a large set "
    "of \"features\" that are generated from the input data. Some of the earliest-used "
    "algorithms, such as decision trees, produced systems of hard if-then rules similar "
    "to the systems of hand-written rules that were then common. Increasingly, however, "
    "research has focused on statistical models, which make soft, probabilistic "
    "decisions based on attaching real-valued weights to each input feature. "
    "Such models have the advantage that they can express the relative certainty of "
    "many different possible answers rather than only one, producing more reliable "
    "results when such a model is included as a component of a larger system."
)


@pytest.fixture(scope="module")
def en_core_web_sm():
    return load_spacy_lang("en_core_web_sm")


@pytest.fixture(scope="module")
def doc(en_core_web_sm):
    return make_spacy_doc((TEXT, {"foo": "bar!"}), lang=en_core_web_sm)


@pytest.fixture(scope="module")
def langs():
    return (
        "en_core_web_sm",
        load_spacy_lang("en_core_web_sm"),
        lambda text: "en_core_web_sm",
    )


class TestLoadSpacyLang:

    @pytest.mark.parametrize("name", ["en_core_web_sm", "es_core_news_sm"])
    def test_load_model(self, name):
        assert isinstance(load_spacy_lang(name), spacy.language.Language)

    @pytest.mark.parametrize("kwargs", [{"exclude": ("tagger", "parser", "ner")}])
    def test_load_model_kwargs(self, kwargs):
        assert isinstance(
            load_spacy_lang("en_core_web_sm", **kwargs),
            spacy.language.Language,
        )

    @pytest.mark.parametrize("kwargs", [{"exclude": ["tagger", "parser", "ner"]}])
    def test_disable_hashability(self, kwargs):
        with pytest.raises(TypeError):
            _ = load_spacy_lang("en_core_web_sm", **kwargs)

    @pytest.mark.parametrize("name", ["en", "un"])
    def test_bad_name(self, name):
        with pytest.raises(OSError):
            _ = load_spacy_lang(name)


class TestMakeSpacyDoc:

    def test_text_data(self, langs):
        text = "This is an English sentence."
        for lang in langs:
            assert isinstance(make_spacy_doc(text, lang=lang), spacy.tokens.Doc)

    def test_record_data(self, langs):
        record = ("This is an English sentence.", {"foo": "bar"})
        for lang in langs:
            assert isinstance(make_spacy_doc(record, lang=lang), spacy.tokens.Doc)

    def test_doc_data(self, langs, en_core_web_sm):
        doc = en_core_web_sm("This is an English sentence.")
        for lang in langs:
            assert isinstance(make_spacy_doc(doc, lang=lang), spacy.tokens.Doc)

    @pytest.mark.parametrize(
        "text, chunk_size",
        [
            # NOTE: I'm cheating with chunk size here, to align with sentence bounds
            ("This is an English sentence. And here is another!", 29),
            ("This is an English sentence. And here is another!", 1000),
        ]
    )
    def test_chunk_size(self, text, chunk_size, en_core_web_sm):
        doc_full = make_spacy_doc(text, en_core_web_sm)
        doc_chunked = make_spacy_doc(text, en_core_web_sm, chunk_size=chunk_size)
        assert isinstance(doc_chunked, spacy.tokens.Doc)
        assert len(doc_full.text) == len(doc_chunked.text)
        assert len(doc_full) == len(doc_chunked)

    @pytest.mark.parametrize(
        "data",
        [
            b"This is an English sentence in bytes.",
            {"content": "This is an English sentence as dict value."},
            True,
        ],
    )
    def test_invalid_data(self, data, en_core_web_sm):
        with pytest.raises(TypeError):
            _ = make_spacy_doc(data, lang=en_core_web_sm)

    @pytest.mark.parametrize(
        "lang", [b"en", ["en_core_web_sm", "es_core_news_sm"], True]
    )
    def test_invalid_lang(self, lang):
        with pytest.raises(TypeError):
            _ = make_spacy_doc("This is an English sentence.", lang=lang)

    @pytest.mark.parametrize(
        "data, lang",
        [
            ("Hello, how are you my friend?", "es_core_news_sm"),
            ("Hello, how are you my friend?", lambda x: "es_core_news_sm"),
        ]
    )
    def test_invalid_data_lang_combo(self, data, lang, en_core_web_sm):
        with pytest.raises((ValueError, TypeError)):
            _ = make_spacy_doc(en_core_web_sm(data), lang=lang)



from textacy import load_spacy_lang
from textacy import cache


def test_cache_clear():
    cache.clear()
    assert len(cache.LRU_CACHE.keys()) == 0


def test_cache_size():
    # check cache size; low thresh but still larger than if the size of
    # loaded data was not being correctly assessed
    # NOTE: should come *after* the function that clears the cache
    _ = load_spacy_lang("en_core_web_sm")
    assert cache.LRU_CACHE.currsize >= 1000



print("import: 'textacy'")
import textacy

print("import: 'textacy.augmentation'")
import textacy.augmentation

print("import: 'textacy.cache'")
import textacy.cache

print("import: 'textacy.constants'")
import textacy.constants

print("import: 'textacy.corpus'")
import textacy.corpus

print("import: 'textacy.datasets'")
import textacy.datasets

print("import: 'textacy.errors'")
import textacy.errors

print("import: 'textacy.extensions'")
import textacy.extensions

print("import: 'textacy.extract'")
import textacy.extract

print("import: 'textacy.io'")
import textacy.io

print("import: 'textacy.lang_id'")
import textacy.lang_id

print("import: 'textacy.preprocessing'")
import textacy.preprocessing

print("import: 'textacy.representations'")
import textacy.representations

print("import: 'textacy.resources'")
import textacy.resources

print("import: 'textacy.similarity'")
import textacy.similarity

print("import: 'textacy.spacier'")
import textacy.spacier

print("import: 'textacy.text_stats'")
import textacy.text_stats

print("import: 'textacy.tm'")
import textacy.tm

print("import: 'textacy.types'")
import textacy.types

print("import: 'textacy.utils'")
import textacy.utils

print("import: 'textacy.viz'")
import textacy.viz




import collections

from textacy import make_spacy_doc
from textacy.augmentation import transforms
from textacy.augmentation import utils as aug_utils

import pytest


@pytest.fixture(scope="module")
def spacy_doc():
    text = (
        "Democrats might know that they stand against Trump's policies, but coming up with their own plan is harder than you think. "
        "For a long time, the party's top echelon has been captive to free trade orthodoxy. "
        "Since Bill Clinton, the theory of the case among the Democratic Party's elite has been the more globalization, the better — with mostly a deaf ear turned to the people and places most badly affected. "
        "Worse, their response to globalization's excesses has been: "
        "Here's a new trade deal, much better than the last one."
    )
    return make_spacy_doc(text, lang="en_core_web_sm")


@pytest.fixture(scope="module")
def aug_toks(spacy_doc):
    return aug_utils.to_aug_toks(spacy_doc)


@pytest.mark.skipif(
    aug_utils.concept_net.filepath is None,
    reason="ConceptNet resource must be downloaded before running tests",
)
class TestSubstituteWordSynonyms:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.substitute_word_synonyms(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.substitute_word_synonyms(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.substitute_word_synonyms(aug_toks, num=num)

    def test_pos(self, aug_toks):
        for pos in ["NOUN", ("NOUN", "VERB", "ADJ", "ADV")]:
            new_aug_toks = transforms.substitute_word_synonyms(
                aug_toks, num=1, pos=pos)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.substitute_word_synonyms(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.substitute_word_synonyms(obj, num=1)


@pytest.mark.skipif(
    aug_utils.concept_net.filepath is None,
    reason="ConceptNet resource must be downloaded before running tests",
)
class TestInsertWordSynonyms:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.insert_word_synonyms(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.insert_word_synonyms(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) > len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.insert_word_synonyms(aug_toks, num=num)

    def test_pos(self, aug_toks):
        for pos in ["NOUN", ("NOUN", "VERB", "ADJ", "ADV")]:
            new_aug_toks = transforms.insert_word_synonyms(
                aug_toks, num=1, pos=pos)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) > len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.insert_word_synonyms(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.insert_word_synonyms(obj, num=1)


class TestSwapWords:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.swap_words(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.swap_words(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.swap_words(aug_toks, num=num)

    def test_pos(self, aug_toks):
        for pos in ["NOUN", ("NOUN", "VERB", "ADJ", "ADV")]:
            new_aug_toks = transforms.swap_words(aug_toks, num=1, pos=pos)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.swap_words(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.swap_words(obj, num=1)


class TestDeleteWords:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.delete_words(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.delete_words(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) < len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.delete_words(aug_toks, num=num)

    def test_pos(self, aug_toks):
        for pos in ["NOUN", ("NOUN", "VERB", "ADJ", "ADV")]:
            new_aug_toks = transforms.delete_words(aug_toks, num=1, pos=pos)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) < len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.delete_words(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.delete_words(obj, num=1)


@pytest.mark.skipif(
    aug_utils.udhr.index is None,
    reason="UDHR dataset must be downloaded before running tests",
)
class TestSubstituteChars:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.substitute_chars(aug_toks, num=num, lang="en")
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        # using higher nums here to prevent the very unlikely case
        # that all characters are substituted by the same character
        for num in [3, 5]:
            new_aug_toks = transforms.substitute_chars(aug_toks, num=num, lang="en")
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )
            assert all(
                len(aug_tok.text) == len(new_aug_tok.text)
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.substitute_chars(aug_toks, num=num, lang="en")

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.substitute_chars(aug_toks, num=num, lang="en")
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.substitute_chars(obj, num=1, lang="en")


@pytest.mark.skipif(
    aug_utils.udhr.index is None,
    reason="UDHR dataset must be downloaded before running tests",
)
class TestInsertChars:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.insert_chars(aug_toks, num=num, lang="en")
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.insert_chars(aug_toks, num=num, lang="en")
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert all(
                (
                    aug_tok.text == new_aug_tok.text or
                    len(aug_tok.text) < len(new_aug_tok.text)
                )
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.insert_chars(aug_toks, num=num, lang="en")

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.insert_chars(aug_toks, num=num, lang="en")
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.insert_chars(obj, num=1, lang="en")


class TestSwapChars:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.swap_chars(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        # using higher nums here to prevent the very unlikely case
        # that all characters are swapped with the same character
        for num in [3, 5]:
            new_aug_toks = transforms.swap_chars(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert any(
                aug_tok.text != new_aug_tok.text
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )
            assert all(
                len(aug_tok.text) == len(new_aug_tok.text)
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.swap_chars(aug_toks, num=num)

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.swap_chars(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.swap_chars(obj, num=1)


class TestDeleteChars:

    def test_noop(self, aug_toks):
        for num in [0, 0.0]:
            new_aug_toks = transforms.delete_chars(aug_toks, num=num)
            for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks):
                assert aug_tok.text == new_aug_tok.text

    def test_num_int(self, aug_toks):
        for num in [1, 3]:
            new_aug_toks = transforms.delete_chars(aug_toks, num=num)
            assert isinstance(new_aug_toks, list)
            assert len(new_aug_toks) == len(aug_toks)
            assert all(
                isinstance(aug_tok, aug_utils.AugTok)
                for aug_tok in new_aug_toks
            )
            assert all(
                (
                    aug_tok.text == new_aug_tok.text or
                    len(aug_tok.text) > len(new_aug_tok.text)
                )
                for aug_tok, new_aug_tok in zip(aug_toks, new_aug_toks)
            )

    def test_num_float(self, aug_toks):
        for num in [0.1, 0.3]:
            _ = transforms.delete_chars(aug_toks, num=num)

    def test_errors(self, aug_toks):
        for num in [-1, 2.0]:
            with pytest.raises(ValueError):
                _ = transforms.delete_chars(aug_toks, num=num)
        for obj in [["foo", "bar"], "foo bar"]:
            with pytest.raises(TypeError):
                _ = transforms.delete_chars(obj, num=1)



import sys

from textacy.preprocessing.resources import (
    RE_EMAIL,
    RE_EMOJI,
    RE_HASHTAG,
    RE_NUMBER,
    RE_PHONE_NUMBER,
    RE_SHORT_URL,
    RE_URL,
)


GOOD_EMAILS = [
    "prettyandsimple@example.com",
    "very.common@example.com",
    "disposable.style.email.with+symbol@example.com",
    "other.email-with-dash@example.com",
    "üñîçøðé@example.com",
    "üñîçøðé@üñîçøðé.com",
    "example@s.solutions",
    "あいうえお@example.com",
]
BAD_EMAILS = [
    "plainaddress",
    "abc.example.com",
    "A@b@c@example.com",
    "john..doe@example.com",
    "john.doe@example..com",
    "#@%^%#$@#$@#.com",
    "@example.com",
    "email.example.com",
    "“email”@example.com",
    "email@example",
    "email@-example.com",
    ".email@example.com",
    "email.@example.com",
    "email@111.222.333.44444",
    "email@[123.123.123.123]",
    "user@[IPv6:2001:db8::1]]",
]

GOOD_EMOJIS = [
    "☀", "⛿",  # miscellaneous symbols
    "✀", "➿",  # dingbats
]
if sys.maxunicode >= 0x10ffff:
    GOOD_EMOJIS.extend([
        "🌀", "🗿",  # miscellaneous symbols and pictographs
        "😀", "🙏",  # emoticons
        "🚀", "🛺",  # transport and map symbols
        "🤀", "🧿",  # supplemental symbols and pictographs
        "🩰", "🪕",  # symbols and pictographs extended-a
    ])

# source: https://github.com/twitter/twitter-text
GOOD_HASHTAGS = [
    "#hashtag",
    "#Azərbaycanca",
    "#mûǁae",
    "#Čeština",
    "#Ċaoiṁín",
    "#עברית",
    "#العربية",
]
BAD_HASHTAGS = [
    "#0",
    "これはダメ#ハッシュタグ",
]

GOOD_NUMBERS = [
    "1",
    "1,234",
    "1.234",
    "1 234",
    "1,234.56",
    "1.234,56",
    "1 234,56",
    "123,456,789",
    "123.456.789",
    "123,456.789",
    "123 456.789",
    "123 456,789",
    "-123",
    "–123",
    "+123",
    "0.123",
    ".123",
    "3.141592653589793238",
    "1,000,000,000,000.00",
]
BAD_NUMBERS = ["D3", "3D", "1st"]
PARTIAL_NUMBERS = [
    "111-111",
    "011,111,111",
    "111.",
    "01,111,111",
    "$123.45",
    "(555)123-456",
    "2015-12-24",
]

GOOD_PHONE_NUMBERS = [
    "1-722-686-3338",
    "984.744.3425",
    "(188)273-7152",
    "+1 (616) 555-3439",
    "(027)458-7382x7531",
    "727.769.2515 #64526",
    "(535) 327 1955 ext. 902",
    "1-492-748-2325-1056",
    "099 145 5237",
    "040.351.7778x63654",
    "123-4567",
]
BAD_PHONE_NUMBERS = [
    "+36(0)4963872475",
    "2015-12-23",
    "12/23/2015",
    "(044) 664 123 45 67",
    "(12) 3456 7890",
    "01234 567890",
    "91 23 45 678",
    "12-345-67-89",
    "123,456,789",
]
PARTIAL_PHONE_NUMBERS = ["(0123) 456 7890"]

GOOD_URLS = [
    "http://foo.com/blah_blah",
    "http://foo.com/blah_blah/",
    "http://foo.com/blah_blah_(wikipedia)",
    "http://foo.com/blah_blah_(wikipedia)_(again)",
    "http://www.example.com/wpstyle/?p=364",
    "https://www.example.com/foo/?bar=baz&inga=42&quux",
    "http://✪df.ws/123",
    "http://userid:password@example.com:8080",
    "http://userid:password@example.com:8080/",
    "http://userid@example.com",
    "http://userid@example.com/",
    "http://userid@example.com:8080",
    "http://userid@example.com:8080/",
    "http://userid:password@example.com",
    "http://userid:password@example.com/",
    "http://142.42.1.1/",
    "http://142.42.1.1:8080/",
    "http://➡.ws/䨹",
    "http://⌘.ws",
    "http://⌘.ws/",
    "http://foo.com/blah_(wikipedia)#cite-1",
    "http://foo.com/blah_(wikipedia)_blah#cite-1",
    "http://foo.com/unicode_(✪)_in_parens",
    "http://foo.com/(something)?after=parens",
    "http://☺.damowmow.com/",
    "http://code.google.com/events/#&product=browser",
    "http://j.mp",
    "ftp://foo.bar/baz",
    "http://foo.bar/?q=Test%20URL-encoded%20stuff",
    "http://مثال.إختبار",
    "http://例子.测试",
    "http://उदाहरण.परीक्षा",
    "http://-.~_!$&'()*+,;=:%40:80%2f::::::@example.com",
    "http://1337.net",
    "http://a.b-c.de",
    "http://223.255.255.254",
    "www.foo.com",
    "www3.foo.com/bar",
]
BAD_URLS = [
    "http://",
    "http://.",
    "http://..",
    "http://../",
    "http://?",
    "http://??",
    "http://??/",
    "http://#",
    "http://##",
    "http://##/",
    "//",
    "//a",
    "///a",
    "///",
    "http:///a",
    "foo.com",
    "rdar://1234",
    "h://test",
    "http:// shouldfail.com",
    ":// should fail",
    "ftps://foo.bar/",
    "http://-error-.invalid/",
    "http://a.b--c.de/",
    "http://-a.b.co",
    "http://a.b-.co",
    "http://0.0.0.0",
    "http://10.1.1.0",
    "http://10.1.1.255",
    "http://224.1.1.1",
    "http://123.123.123",
    "http://3628126748",
    "http://.www.foo.bar/",
    "http://.www.foo.bar./",
    "http://10.1.1.1",
    "foo.bar",
    "can.not.even",
]
PARTIAL_URLS = [
    "http://foo.bar/foo(bar)baz quux",
    "http://1.1.1.1.1",
    "http://foo.bar?q=Spaces should be encoded",
    "http://www.foo.bar./",
]

GOOD_SHORT_URLS = [
    "http://adf.ly/1TxZVO",
    "https://goo.gl/dmL4Gm",
    "http://chart.bt/1OLMAOm",
    "http://ow.ly/Whoc3",
    "http://tinyurl.com/oj6fudq",
    "http://tiny.cc/da1j7x",
    "https://tr.im/KI7ef",
    "http://is.gd/o46NHa",
    "http://yep.it/ywdiux",
    "http://snipurl.com/2adb7eb",
    "http://adyou.me/I4YW",
    "http://nyti.ms/1TgPKgX",
    "http://qr.net/bpfqD",
    "https://chartbeat.com/about/",
    "http://subfoo.foo.bar/abcd",
]
BAD_SHORT_URLS = [
    "ftp://foo.bar/baz",
    "http://foo.com/blah_blah?adsf",
    "https://www.example.com/foo/?bar=baz&inga=42&quux",
]


class TestEmailRegex:

    def test_good_emails(self):
        for item in GOOD_EMAILS:
            assert item == RE_EMAIL.search(item).group()

    def test_bad_emails(self):
        for item in BAD_EMAILS:
            assert RE_EMAIL.search(item) is None


class TestEmojiRegex:

    def test_good_emojis(self):
        for item in GOOD_EMOJIS:
            match = RE_EMOJI.search(item)
            assert match is not None
            assert item == match.group()


class TestHashtagRegex:

    def test_good_hashtags(self):
        for item in GOOD_HASHTAGS:
            assert item == RE_HASHTAG.search(item).group()

    def test_bad_hashtags(self):
        for item in BAD_HASHTAGS:
            assert RE_HASHTAG.search(item) is None


class TestNumberRegex:

    def test_good_numbers(self):
        for item in GOOD_NUMBERS:
            assert item == RE_NUMBER.search(item).group()

    def test_bad_numbers(self):
        for item in BAD_NUMBERS:
            assert RE_NUMBER.search(item) is None

    def test_partial_numbers(self):
        for item in PARTIAL_NUMBERS:
            assert item != RE_NUMBER.search(item)


class TestPhoneNumberRegex:

    def test_good_phone_numbers(self):
        for item in GOOD_PHONE_NUMBERS:
            assert item == RE_PHONE_NUMBER.search(item).group()

    def test_bad_phone_numbers(self):
        for item in BAD_PHONE_NUMBERS:
            assert RE_PHONE_NUMBER.search(item) is None

    def test_partial_phone_numbers(self):
        for item in PARTIAL_PHONE_NUMBERS:
            assert item != RE_PHONE_NUMBER.search(item)


class TestShortURLRegex:

    def test_good_short_urls(self):
        for item in GOOD_SHORT_URLS:
            assert item == RE_SHORT_URL.search(item).group()

    def test_bad_short_urls(self):
        for item in BAD_SHORT_URLS:
            assert RE_SHORT_URL.search(item) is None


class TestURLRegex:

    def test_good_urls(self):
        for item in GOOD_URLS:
            assert item == RE_URL.search(item).group()

    def test_bad_urls(self):
        for item in BAD_URLS:
            assert RE_URL.search(item) is None

    def test_partial_urls(self):
        for item in PARTIAL_URLS:
            assert item != RE_URL.search(item)



"""
Language Identification
-----------------------

:mod:`textacy.lang_id`: Interface for de/serializing a language identification model,
and using it to identify the most probable language(s) of a given text. Inspired by
Google's Compact Language Detector v3 (https://github.com/google/cld3) and
implemented with ``thinc`` v8.0.

Model
^^^^^

Character unigrams, bigrams, and trigrams are extracted separately from the first
1000 characters of lower-cased input text. Each collection of ngrams is hash-embedded
into a 100-dimensional space, then averaged. The resulting feature vectors are
concatenated into a single embedding layer, then passed on to a dense layer with
ReLu activation and finally a Softmax output layer. The model's predictions give
the probabilities for a text to be written in ~140 ISO 639-1 languages.

Dataset
^^^^^^^

The model was trained on a randomized, stratified subset of ~375k texts
drawn from several sources:

- **WiLi:** A public dataset of short text extracts from Wikipedias in over 230
  languages. Style is relatively formal; subject matter is "encyclopedic".
  Source: https://zenodo.org/record/841984
- **Tatoeba:** A crowd-sourced collection of sentences and their translations into
  many languages. Style is relatively informal; subject matter is a variety of
  everyday things and goings-on.
  Source: https://tatoeba.org/eng/downloads.
- **UDHR:** The UN's Universal Declaration of Human Rights document, translated into
  hundreds of languages and split into paragraphs. Style is formal; subject matter is
  fundamental human rights to be universally protected.
  Source: https://unicode.org/udhr/index.html
- **DSLCC**: Two collections of short excerpts of journalistic texts in a handful
  of language groups that are highly similar to each other. Style is relatively formal;
  subject matter is current events.
  Source: http://ttg.uni-saarland.de/resources/DSLCC/

Performance
^^^^^^^^^^^

The trained model achieved F1 = 0.97 when averaged over all languages.

A few languages have worse performance; for example, the two Norwegians ("nb" and "no"),
as well as Bosnian ("bs"), Serbian ("sr"), and Croatian ("hr"), which are extremely
similar to each other. See the textacy-data releases for more details:
https://github.com/bdewilde/textacy-data/releases/tag/lang-identifier-v2.0
"""
from __future__ import annotations

import logging
import pathlib
import urllib
from typing import List, Tuple

from thinc.api import Model

from . import models
from .. import constants, utils


LOGGER = logging.getLogger(__name__)


class LangIdentifier:
    """
    Args:
        version
        data_dir
        model_base

    Attributes:
        model
        classes
    """

    def __init__(
        self,
        version: float | str,
        data_dir: str | pathlib.Path = constants.DEFAULT_DATA_DIR.joinpath("lang_identifier"),
        model_base: Model = models.LangIdentifierModelV2(),
    ):
        self.data_dir = utils.to_path(data_dir)
        self.version = str(version)
        self._model_base = model_base
        self._model = None
        self._classes = None

    @property
    def model_id(self) -> str:
        return f"lang-identifier-v{self.version}"

    @property
    def model_fpath(self) -> pathlib.Path:
        return self.data_dir.joinpath(f"{self.model_id}.bin")

    @property
    def model(self) -> Model:
        if self._model is None:
            self._model = self.load_model()
        return self._model

    @property
    def classes(self):
        if self._classes is None:
            self._classes = self.model.layers[-1].attrs["classes"]
        return self._classes

    def save_model(self):
        """Save trained :attr:`LangIdentifier.model` to disk, as bytes."""
        LOGGER.info("saving LangIdentifier model to %s", self.model_fpath)
        self.model.to_disk(self.model_fpath)

    def load_model(self) -> Model:
        """
        Load trained model from bytes on disk, using :attr:`LangIdentifier.model_base`
        as the framework into which the data is fit.
        """
        try:
            LOGGER.debug("loading LangIdentifier model from %s", self.model_fpath)
            return self._model_base.from_disk(self.model_fpath)
        except FileNotFoundError:
            LOGGER.exception(
                "LangIdentifier model not found at %s -- have you downloaded it yet?",
                self.model_fpath,
            )
            raise

    def download(self, force: bool = False):
        """
        Download version-specific model data as a binary file and save it to disk
        at :attr:`LangIdentifier.model_fpath`.

        Args:
            force: If True, download the model data, even if it already exists on disk
                under :attr:`self.data_dir`; otherwise, don't.
        """
        # hide this import, since we'll only ever need it _once_ (per model version)
        from .. import io as tio

        model_fname = self.model_fpath.name
        url = urllib.parse.urljoin(
            "https://github.com/bdewilde/textacy-data/releases/download/",
            self.model_id + "/" + model_fname,
        )
        tio.utils.download_file(
            url, filename=model_fname, dirpath=self.data_dir, force=force,
        )

    def identify_lang(
        self,
        text: str,
        with_probs: bool = False,
    ) -> str | Tuple[str, float]:
        """
        Identify the most probable language identified in ``text``,
        with or without the corresponding probability.

        Args:
            text
            with_probs

        Returns:
            ISO 639-1 standard language code of the most probable language,
            optionally with its probability.
        """
        if not self._is_valid_text(text):
            result = ("un", 1.0)
        else:
            text_ = utils.to_collection(text, str, list)
            result = models.get_topn_preds_and_probs(
                self.model.predict(text_), 1, self.classes
            )[0][0]
        return result[0] if with_probs is False else result

    def identify_topn_langs(
        self,
        text: str,
        topn: int = 3,
        with_probs: bool = False,
    ) -> List[str] | List[Tuple[str, float]]:
        """
        Identify the ``topn`` most probable languages identified in ``text``,
        with or without the corresponding probabilities.

        Args:
            text
            topn
            with_probs

        Returns:
            ISO 639-1 standard language code and optionally with its probability
            of the ``topn`` most probable languages.
        """
        if not self._is_valid_text(text):
            results = [("un", 1.0)]
        else:
            text_ = utils.to_collection(text, str, list)
            results = models.get_topn_preds_and_probs(
                self.model.predict(text_), topn, self.classes
            )[0]
        return [lang for lang, _ in results] if with_probs is False else results

    def _is_valid_text(self, text: str) -> bool:
        return any(char.isalpha() for char in text)


lang_identifier = LangIdentifier(
    version="2.0",
    data_dir=constants.DEFAULT_DATA_DIR.joinpath("lang_identifier"),
    model_base=models.LangIdentifierModelV2(),
)
# expose this as primary user-facing API
# TODO: there's gotta be a better way, this whole setup feels clunky
identify_lang = lang_identifier.identify_lang
identify_topn_langs = lang_identifier.identify_topn_langs



import pytest

from textacy import extract
from textacy import make_spacy_doc
from textacy.text_stats import basics


@pytest.fixture(scope="module")
def doc():
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice."
    )
    return make_spacy_doc(text, lang="en_core_web_sm")


@pytest.fixture(scope="module")
def words(doc):
    return list(
        extract.words(doc, filter_punct=True, filter_stops=False, filter_nums=False)
    )


def test_n_words(doc, words):
    n_doc = basics.n_words(doc)
    n_words = basics.n_words(words)
    assert n_doc == n_words == pytest.approx(26, rel=0.05)


def test_n_unique_words(doc, words):
    n_doc = basics.n_unique_words(doc)
    n_words = basics.n_unique_words(words)
    assert n_doc == n_words == pytest.approx(25, rel=0.05)


def test_n_chars_per_word(doc, words):
    n_doc = basics.n_chars_per_word(doc)
    n_words = basics.n_chars_per_word(words)
    n_expected = (
        4, 5, 5, 2, 2, 5, 3, 6, 5, 7, 9, 7, 3, 2, 8, 4, 7, 9, 4, 3, 6, 4, 3, 2, 8, 3,
    )
    assert n_doc == n_words == n_expected


def test_n_chars(words):
    n_chars_per_word = basics.n_chars_per_word(words)
    n_chars = basics.n_chars(n_chars_per_word)
    assert n_chars == pytest.approx(126, rel=0.05)


@pytest.mark.parametrize("min_n_chars,n_exp", [(6, 9), (7, 7), (8, 4)])
def test_n_long_words(words, min_n_chars, n_exp):
    n_chars_per_word = basics.n_chars_per_word(words)
    n_long_words = basics.n_long_words(n_chars_per_word, min_n_chars=min_n_chars)
    assert n_long_words == n_exp


# NOTE: you can see how the hack syllable counting stumbles, especially on short words
def test_n_syllables_per_word(doc, words):
    n_doc = basics.n_syllables_per_word(doc, "en")
    n_words = basics.n_syllables_per_word(words, "en")
    n_expected = (
        1, 1, 1, 1, 1, 1, 1, 2, 1, 3, 3, 1, 1, 1, 3, 1, 2, 3, 1, 1, 2, 1, 1, 1, 3, 1
    )
    assert n_doc == n_words == n_expected


def test_n_syllables(words):
    n_syllables_per_word = basics.n_syllables_per_word(words, "en")
    n_syllables = basics.n_syllables(n_syllables_per_word)
    assert n_syllables == pytest.approx(39, rel=0.05)


def test_n_monosyllable_words(words):
    n_syllables_per_word = basics.n_syllables_per_word(words, "en")
    n_monosyllable_words = basics.n_monosyllable_words(n_syllables_per_word)
    assert n_monosyllable_words == pytest.approx(18, rel=0.05)


@pytest.mark.parametrize("min_n_syllables,n_exp", [(2, 8), (3, 5), (4, 0)])
def test_n_polysyllable_words(words, min_n_syllables, n_exp):
    n_syllables_per_word = basics.n_syllables_per_word(words, "en")
    n_polysyllable_words = basics.n_polysyllable_words(
        n_syllables_per_word, min_n_syllables=min_n_syllables,
    )
    assert n_polysyllable_words == n_exp


def test_n_sents(doc):
    n_obs = basics.n_sents(doc)
    assert n_obs == 1


def test_entropy(doc, words):
    val_doc = basics.entropy(doc)
    val_words = basics.entropy(words)
    assert val_doc == pytest.approx(4.623516641218013, abs=1e-4)
    assert val_words == pytest.approx(4.623516641218013, abs=1e-4)



import collections
import itertools
import re

import pytest
from spacy.tokens import Span

import textacy
import textacy.datasets
import textacy.extract
import textacy.preprocessing


pytestmark = pytest.mark.skipif(
    textacy.datasets.CapitolWords().filepath is None,
    reason="CapitolWords dataset must be downloaded before running tests",
)


@pytest.fixture(scope="module")
def dataset():
    dataset = textacy.datasets.CapitolWords()
    return dataset


@pytest.fixture(scope="module")
def record(dataset):
    record = next(dataset.records(limit=1))
    return record


@pytest.fixture(scope="module")
def doc(record):
    doc = textacy.make_spacy_doc(record, lang="en_core_web_sm")
    return doc


@pytest.fixture(scope="module")
def corpus(dataset):
    corpus = textacy.Corpus("en_core_web_sm", data=dataset.records(limit=10))
    return corpus


def test_dataset(dataset):
    assert hasattr(dataset, "info")
    assert isinstance(dataset.info, dict)
    assert hasattr(dataset, "download")


def test_dataset_record(record):
    assert isinstance(record, tuple)
    assert len(record) == 2
    assert isinstance(record[0], str) and isinstance(record[1], dict)
    assert hasattr(record, "text") and hasattr(record, "meta")


def test_extract_kwic(record):
    results = list(
        textacy.extract.keyword_in_context(
            record.text, "work(ing|ers?)", window_width=35
        )
    )
    assert results
    assert len(results) == 4
    assert all(isinstance(result, tuple) for result in results)


def test_preprocessing(record):
    result = textacy.preprocessing.replace.numbers(record.text)
    assert isinstance(result, str)
    assert not any(re.findall(r"\d+", result))


def test_preprocessing_pipeline(record):
    preproc = textacy.preprocessing.make_pipeline(
        textacy.preprocessing.normalize.unicode,
        textacy.preprocessing.normalize.quotation_marks,
        textacy.preprocessing.normalize.whitespace,
    )
    result = preproc(record.text)
    assert result and isinstance(result, str)


def test_doc_extensions(doc):
    assert all(hasattr(doc._, attr) for attr in ["preview", "meta"])
    assert isinstance(doc._.preview, str)
    assert isinstance(doc._.meta, dict)


def test_extract_token_matches(doc):
    patterns = [
        {"POS": {"IN": ["ADJ", "DET"]}, "OP": "+"}, {"ORTH": {"REGEX": "workers?"}}
    ]
    results = list(textacy.extract.token_matches(doc, patterns))
    assert results and len(results) == 2
    assert all(isinstance(result, Span) for result in results)


def test_corpus_metadata(corpus):
    result = corpus.agg_metadata("congress", min)
    assert result and isinstance(result, int)
    min_dt = corpus.agg_metadata("date", min)
    max_dt = corpus.agg_metadata("date", max)
    assert min_dt < max_dt
    result = corpus.agg_metadata("speaker_name", collections.Counter)
    assert result and isinstance(result, collections.Counter)


def test_corpus_extract_token_matches(corpus):
    patterns = [
        {"POS": {"IN": ["ADJ", "DET"]}, "OP": "+"}, {"ORTH": {"REGEX": "workers?"}}
    ]
    matches = itertools.chain.from_iterable(
        textacy.extract.token_matches(doc, patterns) for doc in corpus
    )
    result = collections.Counter(match.lemma_ for match in matches).most_common(10)
    assert all(isinstance(key, str) for key, _ in result)
    assert all(isinstance(val, int) for _, val in result)


def test_doc_extract_keyterms(doc):
    result = doc._.extract_keyterms(
        "textrank", normalize="lemma", window_size=10, edge_weighting="count", topn=10
    )
    assert result and isinstance(result, list)
    assert all(isinstance(key, str) for key, _ in result)
    assert all(isinstance(val, float) for _, val in result)
    assert (
        sorted([val for _, val in result], reverse=True) == [val for _, val in result]
    )



"""
:mod:`textacy.extensions`: Inspect, extend, and transform spaCy's core ``Doc``
data structure, either directly via functions that take a ``Doc`` as their first arg
or as custom attributes / methods on instantiated docs prepended by an underscore:

.. code-block:: pycon

    >>> doc = textacy.make_spacy_doc("This is a short text.", "en_core_web_sm")
    >>> print(get_preview(doc))
    Doc(6 tokens: "This is a short text.")
    >>> print(doc._.preview)
    Doc(6 tokens: "This is a short text.")
"""
from __future__ import annotations

import operator
from typing import Any, Collection, Dict, List, Optional

import cytoolz
from spacy.tokens import Doc

from . import errors, extract, types


def get_preview(doc: Doc) -> str:
    """
    Get a short preview of the ``Doc``, including the number of tokens
    and an initial snippet.
    """
    snippet = doc.text[:50].replace("\n", " ")
    if len(snippet) == 50:
        snippet = snippet[:47] + "..."
    return f'Doc({len(doc)} tokens: "{snippet}")'


def get_meta(doc: Doc) -> dict:
    """Get custom metadata added to ``Doc``."""
    return doc.user_data.get("textacy", {}).get("meta", {})


def set_meta(doc: Doc, value: dict) -> None:
    """Add custom metadata to ``Doc``."""
    if not isinstance(value, dict):
        raise TypeError(errors.type_invalid_msg("value", type(value), Dict))
    try:
        doc.user_data["textacy"]["meta"] = value
    except KeyError:
        # TODO: confirm that this is the same. it is, right??
        doc.user_data["textacy"] = {"meta": value}


def to_tokenized_text(doc: Doc) -> List[List[str]]:
    """
    Transform ``doc`` into an ordered, nested list of token-texts for each sentence.

    Args:
        doc

    Returns:
        A list of tokens' texts for each sentence in ``doc``.

    Note:
        If ``doc`` hasn't been segmented into sentences, the entire document
        is treated as a single sentence.
    """
    if doc.has_annotation("SENT_START"):
        return [[token.text for token in sent] for sent in doc.sents]
    else:
        return [[token.text for token in doc]]


def to_bag_of_words(
    doclike: types.DocLike,
    *,
    by: str = "lemma_",  # Literal["lemma", "lower", "norm", "orth"]
    weighting: str = "count",  # Literal["count", "freq", "binary"]
    **kwargs,
) -> Dict[int, int | float] | Dict[str, int | float]:
    """
    Transform a ``Doc`` or ``Span`` into a bag-of-words: the set of unique words therein
    mapped to their absolute, relative, or binary frequencies of occurrence.

    Args:
        doclike
        by: Attribute by which spaCy ``Token`` s are grouped before counting,
            as given by ``getattr(token, by)``.
            If "lemma", tokens are grouped by their base form w/o inflectional suffixes;
            if "lower", by the lowercase form of the token text;
            if "norm", by the normalized form of the token text;
            if "orth", by the token text exactly as it appears in ``doc``.
            To output keys as strings, simply append an underscore to any of these;
            for example, "lemma_" creates a bag whose keys are token lemmas as strings.
        weighting: Type of weighting to assign to unique words given by ``by``.
            If "count", weights are the absolute number of occurrences (i.e. counts);
            if "freq", weights are counts normalized by the total token count,
            giving their relative frequency of occurrence;
            if "binary", weights are set equal to 1.
        **kwargs: Passed directly on to :func:`textacy.extract.words()`
            - filter_stops: If True, stop words are removed before counting.
            - filter_punct: If True, punctuation tokens are removed before counting.
            - filter_nums: If True, number-like tokens are removed before counting.

    Returns:
        Mapping of a unique word id or string (depending on the value of ``by``)
        to its absolute, relative, or binary frequency of occurrence
        (depending on the value of ``weighting``).

    Note:
        For "freq" weighting, the resulting set of frequencies won't (necessarily) sum
        to 1.0, since all tokens are used when normalizing counts but some (punctuation,
        stop words, etc.) may be filtered out of the bag afterwards.

    See Also:
        :func:`textacy.extract.words()`
    """
    words = extract.words(doclike, **kwargs)
    bow = cytoolz.recipes.countby(operator.attrgetter(by), words)
    bow = _reweight_bag(weighting, bow, doclike)
    return bow


def to_bag_of_terms(
    doclike: types.DocLike,
    *,
    by: str = "lemma_",  # Literal["lemma_", "lemma", "lower_", "lower", "orth_", "orth"]
    weighting: str = "count",  # Literal["count", "freq", "binary"]
    ngs: Optional[int | Collection[int] | types.DocLikeToSpans] = None,
    ents: Optional[bool | types.DocLikeToSpans] = None,
    ncs: Optional[bool | types.DocLikeToSpans] = None,
    dedupe: bool = True,
) -> Dict[str, int] | Dict[str, float]:
    """
    Transform a ``Doc`` or ``Span`` into a bag-of-terms: the set of unique terms therein
    mapped to their absolute, relative, or binary frequencies of occurrence,
    where "terms" may be a combination of n-grams, entities, and/or noun chunks.

    Args:
        doclike
        by: Attribute by which spaCy ``Span`` s are grouped before counting,
            as given by ``getattr(token, by)``.
            If "lemma", tokens are counted by their base form w/o inflectional suffixes;
            if "lower", by the lowercase form of the token text;
            if "orth", by the token text exactly as it appears in ``doc``.
            To output keys as strings, simply append an underscore to any of these;
            for example, "lemma_" creates a bag whose keys are token lemmas as strings.
        weighting: Type of weighting to assign to unique terms given by ``by``.
            If "count", weights are the absolute number of occurrences (i.e. counts);
            if "freq", weights are counts normalized by the total token count,
            giving their relative frequency of occurrence;
            if "binary", weights are set equal to 1.
        ngs: N-gram terms to be extracted.
            If one or multiple ints, :func:`textacy.extract.ngrams(doclike, n=ngs)` is
            used to extract terms; if a callable, ``ngs(doclike)`` is used to extract
            terms; if None, no n-gram terms are extracted.
        ents: Entity terms to be extracted.
            If True, :func:`textacy.extract.entities(doclike)` is used to extract terms;
            if a callable, ``ents(doclike)`` is used to extract terms;
            if None, no entity terms are extracted.
        ncs: Noun chunk terms to be extracted.
            If True, :func:`textacy.extract.noun_chunks(doclike)` is used to extract
            terms; if a callable, ``ncs(doclike)`` is used to extract terms;
            if None, no noun chunk terms are extracted.
        dedupe: If True, deduplicate terms whose spans are extracted by multiple types
            (e.g. a span that is both an n-gram and an entity), as identified by
            identical (start, stop) indexes in ``doclike``; otherwise, don't.

    Returns:
        Mapping of a unique term id or string (depending on the value of ``by``)
        to its absolute, relative, or binary frequency of occurrence
        (depending on the value of ``weighting``).

    See Also:
        :func:`textacy.extract.terms()`
    """
    terms = extract.terms(doclike, ngs=ngs, ents=ents, ncs=ncs, dedupe=dedupe)
    # spaCy made some awkward changes in the Span API, making it harder to get int ids
    # and adding inconsistencies with equivalent Token API
    # so, here are some hacks to spare users that annoyance
    if by.startswith("lower"):
        bot = cytoolz.recipes.countby(lambda span: span.text.lower(), terms)
    else:
        by_ = by if by.endswith("_") else f"{by}_"
        bot = cytoolz.recipes.countby(operator.attrgetter(by_), terms)
    # if needed, here we take the term strings and manually convert back to ints
    if not by.endswith("_"):
        ss = doclike.vocab.strings
        bot = {ss.add(term_str): weight for term_str, weight in bot.items()}
    bot = _reweight_bag(weighting, bot, doclike)
    return bot


def _reweight_bag(
    weighting: str, bag: Dict[Any, int], doclike: types.DocLike
) -> Dict[Any, int] | Dict[Any, float]:
    if weighting == "count":
        return bag
    elif weighting == "freq":
        n_tokens = len(doclike)
        return {term: weight / n_tokens for term, weight in bag.items()}
    elif weighting == "binary":
        return {term: 1 for term in bag.keys()}
    else:
        raise ValueError(
            errors.value_invalid_msg(
                "weighting", weighting, {"count", "freq", "binary"}
            )
        )


def get_doc_extensions() -> Dict[str, Dict[str, Any]]:
    """
    Get textacy's custom property and method doc extensions
    that can be set on or removed from the global :class:`spacy.tokens.Doc`.
    """
    return _DOC_EXTENSIONS


def set_doc_extensions():
    """
    Set textacy's custom property and method doc extensions
    on the global :class:`spacy.tokens.Doc`.
    """
    for name, kwargs in get_doc_extensions().items():
        if not Doc.has_extension(name):
            Doc.set_extension(name, **kwargs)


def remove_doc_extensions():
    """
    Remove textacy's custom property and method doc extensions
    from the global :class:`spacy.tokens.Doc`.
    """
    for name in get_doc_extensions().keys():
        _ = Doc.remove_extension(name)


_DOC_EXTENSIONS: Dict[str, Dict[str, Any]] = {
    # property extensions
    "preview": {"getter": get_preview},
    "meta": {"getter": get_meta, "setter": set_meta},
    # method extensions
    "to_tokenized_text": {"method": to_tokenized_text},
    "to_bag_of_words": {"method": to_bag_of_words},
    "to_bag_of_terms": {"method": to_bag_of_terms},
}



"""
Edit-based Metrics
------------------

:mod:`textacy.similarity.edits`: Normalized similarity metrics built on edit-based
algorithms that compute the number of operations (additions, subtractions, ...)
needed to transform one string into another.
"""
from __future__ import annotations

from typing import Optional

import sklearn.feature_extraction
import sklearn.metrics
from jellyfish import (
    hamming_distance as _hamming,
    levenshtein_distance as _levenshtein,
    jaro_similarity as _jaro_similarity,
)

from .. import constants


def _shortcut(str1: str, str2: str) -> Optional[float]:
    if not str1 and not str2:
        return 0.0
    elif str1 == str2:
        return 1.0
    else:
        return None


def hamming(str1: str, str2: str) -> float:
    """
    Compute the similarity between two strings using Hamming distance,
    which gives the number of characters at corresponding string indices that differ,
    including chars in the longer string that have no correspondents in the shorter.

    Args:
        str1
        str2

    Returns:
        Similarity between ``str1`` and ``str2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings
    """
    shortcut = _shortcut(str1, str2)
    if shortcut:
        return shortcut

    distance = _hamming(str1, str2)
    max_len = max(len(str1), len(str2))
    try:
        return 1.0 - (distance / max_len)
    except ZeroDivisionError:
        return 0.0


def levenshtein(str1: str, str2: str) -> float:
    """
    Measure the similarity between two strings using Levenshtein distance,
    which gives the minimum number of character insertions, deletions, and substitutions
    needed to change one string into the other.

    Args:
        str1
        str2

    Returns:
        Similarity between ``str1`` and ``str2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings
    """
    shortcut = _shortcut(str1, str2)
    if shortcut:
        return shortcut

    distance = _levenshtein(str1, str2)
    max_len = max(len(str1), len(str2))
    try:
        return 1.0 - (distance / max_len)
    except ZeroDivisionError:
        return 0.0


def jaro(str1: str, str2: str) -> float:
    """
    Measure the similarity between two strings using Jaro (*not* Jaro-Winkler) distance,
    which searches for common characters while taking transpositions and string lengths
    into account.

    Args:
        str1
        str2

    Returns:
        Similarity between ``str1`` and ``str2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings
    """
    # NOTE: I usually try to avoid wrapping other packages' functionality in one-liners
    # but this is here for consistency with levenshtein and hamming,
    # which do extra work to get normalized similarity values
    shortcut = _shortcut(str1, str2)
    if shortcut:
        return shortcut

    return _jaro_similarity(str1, str2)


def character_ngrams(str1: str, str2: str) -> float:
    """
    Measure the similarity between two strings using a character ngrams similarity
    metric, in which strings are transformed into trigrams of alnum-only characters,
    vectorized and weighted by tf-idf, then compared by cosine similarity.

    Args:
        str1
        str2

    Returns:
        Similarity between ``str1`` and ``str2`` in the interval [0.0, 1.0],
        where larger values correspond to more similar strings

    Note:
        This method has been used in cross-lingual plagiarism detection and
        authorship attribution, and seems to work better on longer texts.
        At the very least, it is *slow* on shorter texts relative to the other
        similarity measures.
    """
    # TODO: figure out where this method falls -- not edit-based??
    vectorizer = sklearn.feature_extraction.text.TfidfVectorizer(
        preprocessor=lambda s: " ".join(constants.RE_ALNUM.findall(s.lower())),
        analyzer="char",
        token_pattern="(?u)\\b\\w+\\b",
        ngram_range=(3, 3),
    )
    mat = vectorizer.fit_transform([str1, str2])
    return sklearn.metrics.pairwise.cosine_similarity(mat[0, :], mat[1, :]).item()



from __future__ import annotations

import itertools
import operator
import re
from typing import (
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

from cytoolz import itertoolz
from spacy.tokens import Doc, Token

from .. import constants, errors, types, utils
from . import matches


def terms_to_strings(
    terms: Iterable[types.SpanLike], by: str | Callable[[types.SpanLike], str],
) -> Iterable[str]:
    """
    Transform a sequence of terms as spaCy ``Token`` s or ``Span`` s into strings.

    Args:
        terms
        by: Method by which terms are transformed into strings.
            If "orth", terms are represented by their text exactly as written;
            if "lower", by the lowercased form of their text;
            if "lemma", by their base form w/o inflectional suffixes;
            if a callable, must accept a ``Token`` or ``Span`` and return a string.

    Yields:
        Next term in ``terms``, as a string.
    """
    if by == "lower":
        terms = (term.text.lower() for term in terms)
    elif by in ("lemma", "orth"):
        by_ = operator.attrgetter(f"{by}_")
        terms = (by_(term) for term in terms)
    elif callable(by):
        terms = (by(term) for term in terms)
    else:
        raise ValueError(
            errors.value_invalid_msg("by", by, {"orth", "lower", "lemma", Callable})
        )
    for term in terms:
        yield term


def clean_term_strings(terms: Iterable[str]) -> Iterable[str]:
    """
    Clean up a sequence of single- or multi-word terms as strings:
    strip leading/trailing junk chars, handle dangling parens and odd hyphenation,
    and normalize whitespace.

    Args:
        terms: Sequence of terms such as "environment", "plastic pollution",
            or "fossil fuel industry" that may be _unclean_ for whatever reason.

    Yields:
        Next term in ``terms`` but with the cruft cleaned up, excluding terms
        that were entirely cruft

    Warning:
        Terms with (intentionally) unusual punctuation may get "cleaned"
        into a form that changes or obscures the original meaning of the term.
    """
    # get rid of leading/trailing junk characters
    terms = (constants.RE_LEAD_TAIL_CRUFT_TERM.sub("", term) for term in terms)
    terms = (constants.RE_LEAD_HYPHEN_TERM.sub(r"\1", term) for term in terms)
    # handle dangling/backwards parens, don't allow '(' or ')' to appear alone
    terms = (
        ""
        if term.count(")") != term.count("(") or term.find(")") < term.find("(")
        else term
        if "(" not in term
        else constants.RE_DANGLING_PARENS_TERM.sub(r"\1\2\3", term)
        for term in terms
    )
    # handle oddly separated hyphenated words
    terms = (
        term
        if "-" not in term
        else constants.RE_NEG_DIGIT_TERM.sub(
            r"\1\2", constants.RE_WEIRD_HYPHEN_SPACE_TERM.sub(r"\1", term)
        )
        for term in terms
    )
    # handle oddly separated apostrophe'd words
    terms = (
        constants.RE_WEIRD_APOSTR_SPACE_TERM.sub(r"\1\2", term) if "'" in term else term
        for term in terms
    )
    # normalize whitespace
    terms = (constants.RE_NONBREAKING_SPACE.sub(" ", term).strip() for term in terms)
    for term in terms:
        if re.search(r"\w", term):
            yield term


def aggregate_term_variants(
    terms: Set[str],
    *,
    acro_defs: Optional[Dict[str, str]] = None,
    fuzzy_dedupe: bool = True,
) -> List[Set[str]]:
    """
    Take a set of unique terms and aggregate terms that are symbolic, lexical,
    and ordering variants of each other, as well as acronyms and fuzzy string matches.

    Args:
        terms: Set of unique terms with potential duplicates
        acro_defs: If not None, terms that are acronyms will be aggregated
            with their definitions and terms that are definitions will be aggregated
            with their acronyms
        fuzzy_dedupe: If True, fuzzy string matching will be used
            to aggregate similar terms of a sufficient length

    Returns:
        Each item is a set of aggregated terms.

    Notes:
        Partly inspired by aggregation of variants discussed in
        Park, Youngja, Roy J. Byrd, and Branimir K. Boguraev.
        "Automatic glossary extraction: beyond terminology identification."
        Proceedings of the 19th international conference on Computational linguistics-
        Volume 1. Association for Computational Linguistics, 2002.
    """
    from .. import similarity  # ugh, hide import here

    agg_terms = []
    seen_terms: Set[str] = set()
    for term in sorted(terms, key=len, reverse=True):
        if term in seen_terms:
            continue
        variants = set([term])
        seen_terms.add(term)

        # symbolic variations
        if "-" in term:
            variant = term.replace("-", " ").strip()
            if variant in terms.difference(seen_terms):
                variants.add(variant)
                seen_terms.add(variant)
        if "/" in term:
            variant = term.replace("/", " ").strip()
            if variant in terms.difference(seen_terms):
                variants.add(variant)
                seen_terms.add(variant)

        # lexical variations
        term_words = term.split()
        # last_word = term_words[-1]
        # # assume last word is a noun
        # last_word_lemmatized = lemmatizer.lemmatize(last_word, 'n')
        # # if the same, either already a lemmatized noun OR a verb; try verb
        # if last_word_lemmatized == last_word:
        #     last_word_lemmatized = lemmatizer.lemmatize(last_word, 'v')
        # # if at least we have a new term... add it
        # if last_word_lemmatized != last_word:
        #     term_lemmatized = ' '.join(term_words[:-1] + [last_word_lemmatized])
        #     if term_lemmatized in terms.difference(seen_terms):
        #         variants.add(term_lemmatized)
        #         seen_terms.add(term_lemmatized)

        # if term is an acronym, add its definition
        # if term is a definition, add its acronym
        if acro_defs:
            for acro, def_ in acro_defs.items():
                if acro.lower() == term.lower():
                    variants.add(def_.lower())
                    seen_terms.add(def_.lower())
                    break
                elif def_.lower() == term.lower():
                    variants.add(acro.lower())
                    seen_terms.add(acro.lower())
                    break

        # if 3+ -word term differs by one word at the start or the end
        # of a longer phrase, aggregate
        if len(term_words) > 2:
            term_minus_first_word = " ".join(term_words[1:])
            term_minus_last_word = " ".join(term_words[:-1])
            if term_minus_first_word in terms.difference(seen_terms):
                variants.add(term_minus_first_word)
                seen_terms.add(term_minus_first_word)
            if term_minus_last_word in terms.difference(seen_terms):
                variants.add(term_minus_last_word)
                seen_terms.add(term_minus_last_word)
            # check for "X of Y" <=> "Y X" term variants
            if " of " in term:
                split_term = term.split(" of ")
                variant = split_term[1] + " " + split_term[0]
                if variant in terms.difference(seen_terms):
                    variants.add(variant)
                    seen_terms.add(variant)

        # intense de-duping for sufficiently long terms
        if fuzzy_dedupe is True and len(term) >= 13:
            for other_term in sorted(
                terms.difference(seen_terms), key=len, reverse=True
            ):
                if len(other_term) < 13:
                    break
                tsr = similarity.token_sort_ratio(term, other_term)
                if tsr > 0.93:
                    variants.add(other_term)
                    seen_terms.add(other_term)
                    break

        agg_terms.append(variants)

    return agg_terms


def get_longest_subsequence_candidates(
    doc: Doc, match_func: Callable[[Token], bool],
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are longest consecutive
    subsequences of tokens for which all ``match_func(token)`` is True.

    Args:
        doc
        match_func: Function applied sequentially to each ``Token`` in ``doc``
            that returns True for matching ("good") tokens, False otherwise.

    Yields:
        Next longest consecutive subsequence candidate, as a tuple of constituent tokens.
    """
    for key, words_grp in itertools.groupby(doc, key=match_func):
        if key is True:
            yield tuple(words_grp)


def get_ngram_candidates(
    doc: Doc,
    ns: int | Collection[int],
    *,
    include_pos: Optional[str | Collection[str]] = ("NOUN", "PROPN", "ADJ"),
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are n-length sequences
    of tokens (for all n in ``ns``) that don't start/end with a stop word or
    contain punctuation tokens, and whose constituent tokens are filtered by POS tag.

    Args:
        doc
        ns: One or more n values for which to generate n-grams. For example,
            ``2`` gets bigrams; ``(2, 3)`` gets bigrams and trigrams.
        include_pos: One or more POS tags with which to filter ngrams.
            If None, include tokens of all POS tags.

    Yields:
        Next ngram candidate, as a tuple of constituent Tokens.

    See Also:
        :func:`textacy.extract.ngrams()`
    """
    ns = utils.to_collection(ns, int, tuple)
    include_pos = utils.to_collection(include_pos, str, set)
    ngrams = itertoolz.concat(itertoolz.sliding_window(n, doc) for n in ns)
    ngrams = (
        ngram
        for ngram in ngrams
        if not (ngram[0].is_stop or ngram[-1].is_stop)
        and not any(word.is_punct or word.is_space for word in ngram)
    )
    if include_pos:
        ngrams = (
            ngram for ngram in ngrams if all(word.pos_ in include_pos for word in ngram)
        )
    for ngram in ngrams:
        yield ngram


def get_pattern_matching_candidates(
    doc: Doc, patterns: str | List[str] | List[dict] | List[List[dict]],
) -> Iterable[Tuple[Token, ...]]:
    """
    Get candidate keyterms from ``doc``, where candidates are sequences of tokens
    that match any pattern in ``patterns``

    Args:
        doc
        patterns: One or multiple patterns to match against ``doc`` using
            a :class:`spacy.matcher.Matcher`.

    Yields:
        Tuple[:class:`spacy.tokens.Token`]: Next pattern-matching candidate,
        as a tuple of constituent Tokens.

    See Also:
        :func:`textacy.extract.token_matches()`
    """
    for match in matches.token_matches(doc, patterns, on_match=None):
        yield tuple(match)


def get_filtered_topn_terms(
    term_scores: Iterable[Tuple[str, float]],
    topn: int,
    *,
    match_threshold: Optional[float] = None,
) -> List[Tuple[str, float]]:
    """
    Build up a list of the ``topn`` terms, filtering out any that are substrings
    of better-scoring terms and optionally filtering out any that are sufficiently
    similar to better-scoring terms.

    Args:
        term_scores: Iterable of (term, score) pairs, sorted in order of score
            from best to worst. Note that this may be from high to low value or
            low to high, depending on the scoring algorithm.
        topn: Maximum number of top-scoring terms to get.
        match_threshold: Minimal edit distance between a term and previously seen terms,
            used to filter out terms that are sufficiently similar
            to higher-scoring terms. Uses :func:`textacy.similarity.token_sort_ratio()`.
    """
    from .. import similarity  # ugh, hide import here

    topn_terms = []
    seen_terms: Set[str] = set()
    sim_func = similarity.token_sort_ratio
    for term, score in term_scores:
        # skip terms that are substrings of any higher-scoring term
        if any(term in st for st in seen_terms):
            continue
        # skip terms that are sufficiently similar to any higher-scoring term
        if match_threshold and any(
            sim_func(term, st) >= match_threshold for st in seen_terms
        ):
            continue
        seen_terms.add(term)
        topn_terms.append((term, score))
        if len(topn_terms) >= topn:
            break
    return topn_terms


# TODO: decide if we want to keep this function, and if so, where to put it

# import math
# from operator import itemgetter
# from decimal import Decimal

# import numpy as np

# from .. import vsm


# def most_discriminating_terms(
#     terms_lists: Iterable[Iterable[str]],
#     bool_array_grp1: Iterable[bool],
#     *,
#     max_n_terms: int = 1000,
#     top_n_terms: int | float = 25,
# ) -> Tuple[List[str], List[str]]:
#     """
#     Given a collection of documents assigned to 1 of 2 exclusive groups, get the
#     ``top_n_terms`` most discriminating terms for group1-and-not-group2 and
#     group2-and-not-group1.

#     Args:
#         terms_lists: Sequence of documents, each as a sequence of (str) terms;
#             used as input to :func:`doc_term_matrix()`
#         bool_array_grp1: Ordered sequence of True/False values,
#             where True corresponds to documents falling into "group 1" and False
#             corresponds to those in "group 2".
#         max_n_terms: Only consider terms whose document frequency is within
#             the top ``max_n_terms`` out of all distinct terms; must be > 0.
#         top_n_terms: If int (must be > 0), the total number of most discriminating terms
#             to return for each group; if float (must be in the interval (0, 1)),
#             the fraction of ``max_n_terms`` to return for each group.

#     Returns:
#         List of the top ``top_n_terms`` most discriminating terms for grp1-not-grp2, and
#         list of the top ``top_n_terms`` most discriminating terms for grp2-not-grp1.

#     References:
#         King, Gary, Patrick Lam, and Margaret Roberts. "Computer-Assisted Keyword
#         and Document Set Discovery from Unstructured Text." (2014).
#         http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.458.1445&rep=rep1&type=pdf
#     """
#     alpha_grp1 = 1
#     alpha_grp2 = 1
#     if isinstance(top_n_terms, float):
#         top_n_terms = int(round(top_n_terms * max_n_terms))
#     bool_array_grp1 = np.array(bool_array_grp1)
#     bool_array_grp2 = np.invert(bool_array_grp1)

#     vectorizer = vsm.Vectorizer(
#         tf_type="linear",
#         norm=None,
#         idf_type="smooth",
#         min_df=3,
#         max_df=0.95,
#         max_n_terms=max_n_terms,
#     )
#     dtm = vectorizer.fit_transform(terms_lists)
#     id2term = vectorizer.id_to_term

#     # get doc freqs for all terms in grp1 documents
#     dtm_grp1 = dtm[bool_array_grp1, :]
#     n_docs_grp1 = dtm_grp1.shape[0]
#     doc_freqs_grp1 = vsm.get_doc_freqs(dtm_grp1)

#     # get doc freqs for all terms in grp2 documents
#     dtm_grp2 = dtm[bool_array_grp2, :]
#     n_docs_grp2 = dtm_grp2.shape[0]
#     doc_freqs_grp2 = vsm.get_doc_freqs(dtm_grp2)

#     # get terms that occur in a larger fraction of grp1 docs than grp2 docs
#     term_ids_grp1 = np.where(
#         doc_freqs_grp1 / n_docs_grp1 > doc_freqs_grp2 / n_docs_grp2
#     )[0]

#     # get terms that occur in a larger fraction of grp2 docs than grp1 docs
#     term_ids_grp2 = np.where(
#         doc_freqs_grp1 / n_docs_grp1 < doc_freqs_grp2 / n_docs_grp2
#     )[0]

#     # get grp1 terms doc freqs in and not-in grp1 and grp2 docs, plus marginal totals
#     grp1_terms_grp1_df = doc_freqs_grp1[term_ids_grp1]
#     grp1_terms_grp2_df = doc_freqs_grp2[term_ids_grp1]
#     # grp1_terms_grp1_not_df = n_docs_grp1 - grp1_terms_grp1_df
#     # grp1_terms_grp2_not_df = n_docs_grp2 - grp1_terms_grp2_df
#     # grp1_terms_total_df = grp1_terms_grp1_df + grp1_terms_grp2_df
#     # grp1_terms_total_not_df = grp1_terms_grp1_not_df + grp1_terms_grp2_not_df

#     # get grp2 terms doc freqs in and not-in grp2 and grp1 docs, plus marginal totals
#     grp2_terms_grp2_df = doc_freqs_grp2[term_ids_grp2]
#     grp2_terms_grp1_df = doc_freqs_grp1[term_ids_grp2]
#     # grp2_terms_grp2_not_df = n_docs_grp2 - grp2_terms_grp2_df
#     # grp2_terms_grp1_not_df = n_docs_grp1 - grp2_terms_grp1_df
#     # grp2_terms_total_df = grp2_terms_grp2_df + grp2_terms_grp1_df
#     # grp2_terms_total_not_df = grp2_terms_grp2_not_df + grp2_terms_grp1_not_df

#     # get grp1 terms likelihoods, then sort for most discriminating grp1-not-grp2 terms
#     grp1_terms_likelihoods = {}
#     for idx, term_id in enumerate(term_ids_grp1):
#         term1 = (
#             Decimal(math.factorial(grp1_terms_grp1_df[idx] + alpha_grp1 - 1))
#             * Decimal(math.factorial(grp1_terms_grp2_df[idx] + alpha_grp2 - 1))
#             / Decimal(
#                 math.factorial(
#                     grp1_terms_grp1_df[idx]
#                     + grp1_terms_grp2_df[idx]
#                     + alpha_grp1
#                     + alpha_grp2
#                     - 1
#                 )
#             )
#         )
#         term2 = (
#             Decimal(
#                 math.factorial(n_docs_grp1 - grp1_terms_grp1_df[idx] + alpha_grp1 - 1)
#             )
#             * Decimal(
#                 math.factorial(n_docs_grp2 - grp1_terms_grp2_df[idx] + alpha_grp2 - 1)
#             )
#             / Decimal(
#                 (
#                     math.factorial(
#                         n_docs_grp1
#                         + n_docs_grp2
#                         - grp1_terms_grp1_df[idx]
#                         - grp1_terms_grp2_df[idx]
#                         + alpha_grp1
#                         + alpha_grp2
#                         - 1
#                     )
#                 )
#             )
#         )
#         grp1_terms_likelihoods[id2term[term_id]] = term1 * term2
#     top_grp1_terms = [
#         term
#         for term, likelihood in sorted(
#             grp1_terms_likelihoods.items(), key=itemgetter(1), reverse=True
#         )[:top_n_terms]
#     ]

#     # get grp2 terms likelihoods, then sort for most discriminating grp2-not-grp1 terms
#     grp2_terms_likelihoods = {}
#     for idx, term_id in enumerate(term_ids_grp2):
#         term1 = (
#             Decimal(math.factorial(grp2_terms_grp2_df[idx] + alpha_grp2 - 1))
#             * Decimal(math.factorial(grp2_terms_grp1_df[idx] + alpha_grp1 - 1))
#             / Decimal(
#                 math.factorial(
#                     grp2_terms_grp2_df[idx]
#                     + grp2_terms_grp1_df[idx]
#                     + alpha_grp2
#                     + alpha_grp1
#                     - 1
#                 )
#             )
#         )
#         term2 = (
#             Decimal(
#                 math.factorial(n_docs_grp2 - grp2_terms_grp2_df[idx] + alpha_grp2 - 1)
#             )
#             * Decimal(
#                 math.factorial(n_docs_grp1 - grp2_terms_grp1_df[idx] + alpha_grp1 - 1)
#             )
#             / Decimal(
#                 (
#                     math.factorial(
#                         n_docs_grp2
#                         + n_docs_grp1
#                         - grp2_terms_grp2_df[idx]
#                         - grp2_terms_grp1_df[idx]
#                         + alpha_grp2
#                         + alpha_grp1
#                         - 1
#                     )
#                 )
#             )
#         )
#         grp2_terms_likelihoods[id2term[term_id]] = term1 * term2
#     top_grp2_terms = [
#         term
#         for term, likelihood in sorted(
#             grp2_terms_likelihoods.items(), key=itemgetter(1), reverse=True
#         )[:top_n_terms]
#     ]

#     return (top_grp1_terms, top_grp2_terms)



"""
Matches
-------

:mod:`textacy.extract.matches`: Extract matching spans from a document or sentence
using spaCy's built-in matcher or regular expressions.
"""
from __future__ import annotations

import re
from typing import Callable, Dict, Iterable, List, Optional, Pattern, Union

from spacy.matcher import Matcher
from spacy.tokens import Span

from .. import constants, errors, types


def token_matches(
    doclike: types.DocLike,
    patterns: str | List[str] | List[Dict[str, str]] | List[List[Dict[str, str]]],
    *,
    on_match: Optional[Callable] = None,
) -> Iterable[Span]:
    """
    Extract ``Span`` s from a document or sentence matching one or more patterns
    of per-token attr:value pairs, with optional quantity qualifiers.

    Args:
        doclike
        patterns:
            One or multiple patterns to match against ``doclike``
            using a :class:`spacy.matcher.Matcher`.

            If List[dict] or List[List[dict]], each pattern is specified
            as attr: value pairs per token, with optional quantity qualifiers:

            - ``[{"POS": "NOUN"}]`` matches singular or plural nouns,
              like "friend" or "enemies"
            - ``[{"POS": "PREP"}, {"POS": "DET", "OP": "?"}, {"POS": "ADJ", "OP": "?"}, {"POS": "NOUN", "OP": "+"}]``
              matches prepositional phrases, like "in the future" or "from the distant past"
            - ``[{"IS_DIGIT": True}, {"TAG": "NNS"}]`` matches numbered plural nouns,
              like "60 seconds" or "2 beers"
            - ``[{"POS": "PROPN", "OP": "+"}, {}]`` matches proper nouns and
              whatever word follows them, like "Burton DeWilde yaaasss"

            If str or List[str], each pattern is specified as one or more
            per-token patterns separated by whitespace where attribute, value,
            and optional quantity qualifiers are delimited by colons. Note that
            boolean and integer values have special syntax --- "bool(val)" and
            "int(val)", respectively --- and that wildcard tokens still need
            a colon between the (empty) attribute and value strings.

            - ``"POS:NOUN"`` matches singular or plural nouns
            - ``"POS:PREP POS:DET:? POS:ADJ:? POS:NOUN:+"`` matches prepositional phrases
            - ``"IS_DIGIT:bool(True) TAG:NNS"`` matches numbered plural nouns
            - ``"POS:PROPN:+ :"`` matches proper nouns and whatever word follows them

            Also note that these pattern strings don't support spaCy v2.1's
            "extended" pattern syntax; if you need such complex patterns, it's
            probably better to use a List[dict] or List[List[dict]], anyway.

        on_match: Callback function to act on matches.
            Takes the arguments ``matcher``, ``doclike``, ``i`` and ``matches``.

    Yields:
        Next matching ``Span`` in ``doclike``, in order of appearance

    Raises:
        TypeError
        ValueError

    See Also:
        - https://spacy.io/usage/rule-based-matching
        - https://spacy.io/api/matcher
    """  # noqa: E501
    if isinstance(patterns, str):
        patterns = [_make_pattern_from_string(patterns)]
    elif isinstance(patterns, (list, tuple)):
        if all(isinstance(item, str) for item in patterns):
            patterns = [_make_pattern_from_string(pattern) for pattern in patterns]
        elif all(isinstance(item, dict) for item in patterns):
            patterns = [patterns]
        elif all(isinstance(item, (list, tuple)) for item in patterns):
            pass  # already in the right format!
        else:
            raise TypeError(
                errors.type_invalid_msg(
                    "patterns",
                    type(patterns),
                    Union[
                        str, List[str], List[Dict[str, str]], List[List[Dict[str, str]]]
                    ],
                )
            )
    else:
        raise TypeError(
            errors.type_invalid_msg(
                "patterns",
                type(patterns),
                Union[str, List[str], List[Dict[str, str]], List[List[Dict[str, str]]]],
            )
        )
    matcher = Matcher(doclike.vocab)
    matcher.add("match", patterns, on_match=on_match)
    for match in matcher(doclike, as_spans=True):
        yield match


def _make_pattern_from_string(patstr: str) -> List[Dict[str, str]]:
    pattern = []
    for tokpatstr in constants.RE_MATCHER_TOKPAT_DELIM.split(patstr):
        parts = tokpatstr.split(":")
        if 2 <= len(parts) <= 3:
            attr = parts[0]
            attr_val = parts[1]
            if attr and attr_val:
                # handle special bool and int attribute values
                special_val = constants.RE_MATCHER_SPECIAL_VAL.match(attr_val)
                if special_val:
                    attr_val = eval(special_val.group(0))
                tokpat = {attr: attr_val}
            # handle wildcard tokens
            else:
                tokpat = {}
            # handle quantifier ops
            try:
                op_val = parts[2]
                if op_val in constants.MATCHER_VALID_OPS:
                    tokpat["OP"] = op_val
                else:
                    raise ValueError(
                        errors.value_invalid_msg(
                            "op", op_val, constants.MATCHER_VALID_OPS
                        )
                    )
            except IndexError:
                pass
            pattern.append(tokpat)
        else:
            raise ValueError(
                f"pattern string '{patstr}' is invalid; "
                "each element in a pattern string must contain an attribute, "
                "a corresponding value, and an optional quantity qualifier, "
                "delimited by colons, like attr:value:op"
            )
    return pattern


def regex_matches(
    doclike: types.DocLike,
    pattern: str | Pattern,
    *,
    alignment_mode: str = "strict",  # Literal["strict", "contract", "expand"]
) -> Iterable[Span]:
    """
    Extract ``Span`` s from a document or sentence whose full texts match against
    a regular expression ``pattern``.

    Args:
        doclike
        pattern: Valid regular expression against which to match document text,
            either as a string or compiled pattern object.
        alignment_mode: How character indices of regex matches snap to spaCy token
            boundaries. If "strict", only exact alignments are included (no snapping);
            if "contract", tokens completely within the character span are included;
            if "expand", tokens at least partially covered by the character span
            are included.

    Yields:
        Next matching ``Span``.
    """
    for match in re.finditer(pattern, doclike.text):
        start_char_idx, end_char_idx = match.span()
        span = doclike.char_span(
            start_char_idx, end_char_idx, alignment_mode=alignment_mode
        )
        # Doc.char_span() returns None if character indices don’t map to a valid span
        if span is not None:
            yield span



import pytest

import spacy

import textacy


@pytest.fixture(scope="module")
def doc():
    lang = textacy.load_spacy_lang("en_core_web_sm")
    text = (
        "Many years later, as he faced the firing squad, Colonel Aureliano Buendía was "
        "to remember that distant afternoon when his father took him to discover ice. "
        "At that time Macondo was a village of twenty adobe houses, built on the bank "
        "of a river of clear water that ran along a bed of polished stones, which were "
        "white and enormous, like prehistoric eggs. The world was so recent that many "
        "things lacked names, and in order to indicate them it was necessary to point."
    )
    meta = {"author": "Gabriel García Márquez", "title": "Cien años de soledad"}
    return textacy.make_spacy_doc((text, meta), lang=lang)


def test_extensions_exist(doc):
    for name in textacy.get_doc_extensions().keys():
        assert doc.has_extension(name)


def test_set_remove_extensions(doc):
    textacy.remove_doc_extensions()
    for name in textacy.get_doc_extensions().keys():
        assert doc.has_extension(name) is False
    textacy.set_doc_extensions()
    for name in textacy.get_doc_extensions().keys():
        assert spacy.tokens.Doc.has_extension(name) is True


def test_preview(doc):
    preview = doc._.preview
    assert isinstance(preview, str)
    assert preview.startswith("Doc")


class TestMeta:

    def test_getter(self, doc):
        meta = doc._.meta
        assert meta
        assert isinstance(meta, dict)

    def test_setter(self, doc):
        meta = {"foo": "bar"}
        doc._.meta = meta
        assert doc._.meta == meta

    def test_setter_invalid(self, doc):
        with pytest.raises(TypeError):
            doc._.meta = None


def test_tokenized_text(doc):
    result = doc._.to_tokenized_text()
    assert result
    assert (
        isinstance(result, list) and
        isinstance(result[0], list) and
        isinstance(result[0][0], str)
    )
    assert len(result) == sum(1 for _ in doc.sents)


class TestBagOfWords():

    def test_default(self, doc):
        result = doc._.to_bag_of_words()
        assert isinstance(result, dict)
        assert all(
            isinstance(key, str) and isinstance(val, int) for key, val in result.items()
        )

    @pytest.mark.parametrize(
        "by", ["lemma_", "lemma", "lower_", "lower", "norm_", "norm", "orth_", "orth"]
    )
    def test_by(self, by, doc):
        result = doc._.to_bag_of_words(by=by)
        assert isinstance(result, dict)
        if by.endswith("_"):
            assert all(isinstance(key, str) for key in result.keys())
        else:
            assert all(isinstance(key, int) for key in result.keys())

    @pytest.mark.parametrize("weighting", ["count", "freq", "binary"])
    def test_weighting(self, weighting, doc):
        result = doc._.to_bag_of_words(weighting=weighting)
        assert isinstance(result, dict)
        if weighting == "freq":
            assert all(isinstance(val, float) for val in result.values())
        else:
            assert all(isinstance(val, int) for val in result.values())

    @pytest.mark.parametrize(
        "kwargs",
        [{"filter_stops": True}, {"filter_punct": True}, {"filter_nums": True}],
    )
    def test_kwargs(self, kwargs, doc):
        result = doc._.to_bag_of_words(**kwargs)
        assert isinstance(result, dict)

    @pytest.mark.parametrize("by", ["LEMMA", spacy.attrs.LEMMA, True])
    def test_invalid_by(self, by, doc):
        with pytest.raises((AttributeError, TypeError)):
            _ = doc._.to_bag_of_words(by=by)

    @pytest.mark.parametrize("weighting", ["COUNT", "frequency", True])
    def test_invalid_weighting(self, weighting, doc):
        with pytest.raises(ValueError):
            _ = doc._.to_bag_of_words(weighting=weighting)


class TestBagOfTerms():

    def test_all_null(self, doc):
        with pytest.raises(ValueError):
            _ = doc._.to_bag_of_terms()

    @pytest.mark.parametrize(
        "kwargs",
        [{"ngs": 2}, {"ngs": [2, 3], "ents": True}, {"ents": True, "ncs": True}],
    )
    def test_simple_kwargs(self, kwargs, doc):
        result = doc._.to_bag_of_terms(**kwargs)
        assert isinstance(result, dict)
        assert all(
            isinstance(key, str) and isinstance(val, int) for key, val in result.items()
        )

    @pytest.mark.parametrize(
        "by", ["lemma_", "lemma", "lower_", "lower", "orth_", "orth"]
    )
    def test_by(self, by, doc):
        result = doc._.to_bag_of_terms(by=by, ngs=2)
        assert isinstance(result, dict)
        if by.endswith("_"):
            assert all(isinstance(key, str) for key in result.keys())
        else:
            assert all(isinstance(key, int) for key in result.keys())

    @pytest.mark.parametrize("weighting", ["count", "freq", "binary"])
    def test_weighting(self, weighting, doc):
        result = doc._.to_bag_of_terms(weighting=weighting, ngs=2)
        assert isinstance(result, dict)
        if weighting == "freq":
            assert all(isinstance(val, float) for val in result.values())
        else:
            assert all(isinstance(val, int) for val in result.values())

    @pytest.mark.parametrize("by", ["LEMMA", spacy.attrs.LEMMA, True])
    def test_invalid_by(self, by, doc):
        with pytest.raises((AttributeError, TypeError)):
            _ = doc._.to_bag_of_terms(by=by, ngs=2)

    @pytest.mark.parametrize("weighting", ["COUNT", "frequency", True])
    def test_invalid_weighting(self, weighting, doc):
        with pytest.raises(ValueError):
            _ = doc._.to_bag_of_terms(weighting=weighting, ngs=2)
